# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet serializer for App_Io

from typing import Any, final

from pydantic import Field
from overrides import override

from .. import common
from .. import utils


@final
class Command(common.CommandBase):
    """Dataclass for App_Io command

    Attributes
    ----------
    di_to_change: common.FixedList[bool]
        To enable modification on a specific digital interface, set True
    di_state: common.FixedList[bool]
        Output status for each digital interfaces
    """

    di_to_change: common.FixedList[bool] = Field(
        default=common.FixedList[bool](
            12, [True, True, True, True, True, True, True, True, True, True, True, True]
        )
    )
    di_state: common.FixedList[bool] = Field(
        default=common.FixedList[bool](
            12,
            [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        )
    )

    @override
    def is_valid(self) -> bool:
        """Check if the command content is valid or not

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Overridden
        """

        return 0x00 <= self.destination_logical_id <= 0x78


@final
class CommandSerializer(common.CommandSerializerBase):
    """Command serializer for App_Io"""

    @staticmethod
    @override
    def serialize(command: common.SomeCommand) -> common.BarePacket | None:
        """Serialize the given command

        Parameters
        ----------
        command : common.SomeCommand
            App_Io command to serialize

        Returns
        -------
        common.BarePacket | None
            Serialized bytes and its LRC checksum (8bit) if valid

        Notes
        -----
        Static overridden method
        """

        if not (isinstance(command, Command) and command.is_valid()):
            return None

        payload_to_build: bytearray = bytearray()

        payload_to_build.append(min(command.destination_logical_id, 0x78))
        payload_to_build.append(0x80)
        payload_to_build.append(0x01)
        u16_di_state: int = sum(
            (1 if command.di_state[port] else 0) << port for port in range(12)
        )
        u16_di_to_change: int = sum(
            (1 if command.di_to_change[port] else 0) << port for port in range(12)
        )
        payload_to_build.append((u16_di_state >> 8) & 0xFF)
        payload_to_build.append((u16_di_state >> 0) & 0xFF)
        payload_to_build.append((u16_di_to_change >> 8) & 0xFF)
        payload_to_build.append((u16_di_to_change >> 0) & 0xFF)
        payload_to_build.extend([0, 0, 0, 0, 0, 0, 0, 0])

        serialized_packet: dict[str, Any] = {
            "payload": bytes(payload_to_build),
            "checksum": utils.lrc8(bytes(payload_to_build)),
        }

        return common.BarePacket(**serialized_packet)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet parser for App_PAL (MOT, accel)

from datetime import datetime
from typing import Any, final

from overrides import override
from pydantic import Field, field_validator

from .. import common


@final
class ParsedPacket(common.ParsedPacketBase):
    """Dataclass for parsed packets from App_PAL (MOT)

    Attributes
    ----------
    router_serial_id: UInt32
        Serial ID for the first router device (0x80000000 with no routing)
    ai1_voltage: common.UInt16
        Voltage for AI1 port in mV
    sample_count: common.UInt8
        Number of accel samples
    samples_x: common.TimeSeries[common.Int16]
        Accel samples for x axis
    samples_y: common.TimeSeries[common.Int16]
        Accel samples for y axis
    samples_z: common.TimeSeries[common.Int16]
        Accel samples for z axis
    sampling_frequency: UInt16
        Sampling frequency in Hz
    """

    router_serial_id: common.UInt32 = Field(
        default=common.UInt32(0), ge=common.UInt32(0), le=common.UInt32(0xFFFFFFFF)
    )
    ai1_voltage: common.UInt16 = Field(
        default=common.UInt16(0), ge=common.UInt16(0), le=common.UInt16(3700)
    )
    sample_count: common.UInt8 = Field(
        default=common.UInt8(16), ge=common.UInt8(16), le=common.UInt8(16)
    )
    samples_x: common.TimeSeries[common.Int16] = Field(
        default=common.TimeSeries[common.Int16](
            16, [common.Int16(0) for _ in range(16)]
        )
    )
    samples_y: common.TimeSeries[common.Int16] = Field(
        default=common.TimeSeries[common.Int16](
            16, [common.Int16(0) for _ in range(16)]
        )
    )
    samples_z: common.TimeSeries[common.Int16] = Field(
        default=common.TimeSeries[common.Int16](
            16, [common.Int16(0) for _ in range(16)]
        )
    )
    sampling_frequency: common.UInt16 = Field(default=common.UInt16(25))

    @field_validator("sampling_frequency")
    @classmethod
    def check_sampling_frequency(cls, freq: int) -> int:
        """Check for sampling_frequency

        Parameters
        ----------
        freq : int
            Input

        Returns
        -------
        int
            Valid input

        Raises
        ------
        ValueError
            Raise if the specified frequency is not supported
        """

        if freq not in set(
            [
                25,
                50,
                100,
                190,
                380,
                750,
                1100,
                1300,
            ]
        ):
            raise ValueError("not supported")
        return freq


@final
class PacketParser(common.PacketParserBase):
    """Packet parser for App_PAL (MOT)"""

    @staticmethod
    @override
    def is_valid(bare_packet: common.BarePacket) -> bool:
        """Check the given bare packet is valid or not

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Static overridden method
        """
        if (
            (bare_packet.u8_at(0) & 0x80) == 0x80
            and (bare_packet.u8_at(7) & 0x80) == 0x80
            and bare_packet.u8_at(12) == 0x80
            and bare_packet.u8_at(13) == 0x83
            and len(bare_packet.payload) == 188
        ):
            return True
        return False

    @staticmethod
    @override
    def parse(bare_packet: common.BarePacket) -> ParsedPacket | None:
        """Try to parse the given bare packet

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        ParsedPacket | None
            Parsed packet data if valid else None

        Notes
        -----
        Static overridden method
        """
        if not PacketParser.is_valid(bare_packet):
            return None
        frequencies = common.FixedTuple[common.UInt16](
            8,
            [
                common.UInt16(25),
                common.UInt16(50),
                common.UInt16(100),
                common.UInt16(190),
                common.UInt16(380),
                common.UInt16(750),
                common.UInt16(1100),
                common.UInt16(1300),
            ],
        )
        parsed_packet_data: dict[str, Any] = {
            "time_parsed": datetime.now(common.Timezone),
            "packet_type": common.PacketType.APP_PAL_MOT,
            "sequence_number": bare_packet.u16_at(5),
            "source_serial_id": bare_packet.u32_at(7),
            "source_logical_id": bare_packet.u8_at(11),
            "lqi": bare_packet.u8_at(4),
            "supply_voltage": bare_packet.u16_at(19),
            "router_serial_id": bare_packet.u32_at(0),
            "ai1_voltage": bare_packet.u16_at(25),
            "sample_count": 16,
            "samples_x": common.TimeSeries[common.Int16](
                16,
                [bare_packet.i16_at(31 + (10 * index) + 0) for index in range(16)],
            ),
            "samples_y": common.TimeSeries[common.Int16](
                16,
                [bare_packet.i16_at(31 + (10 * index) + 2) for index in range(16)],
            ),
            "samples_z": common.TimeSeries[common.Int16](
                16,
                [bare_packet.i16_at(31 + (10 * index) + 4) for index in range(16)],
            ),
            "sampling_frequency": frequencies[bare_packet.u8_at(29) >> 5],
        }
        return ParsedPacket(**parsed_packet_data)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Utils for MWings

import sys
import os
import subprocess
import time
import math
import re
from pathlib import Path
from typing import Callable

import serial  # type: ignore
from serial.tools import list_ports  # type: ignore


def open_on_system(path: Path) -> None:
    """Open the file in the given path on system using default application

    Parameters
    ----------
    path : Path
        pathlib.Path for file
    """

    if sys.platform == "win32":
        os.startfile(path)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, path])


def ask_user(
    prompt: str,
    regex: str,
    on_error: str,
    ex_verifier: Callable[[str], bool] | None = None,
    max_attempts: int | None = None,
) -> str:
    """Ask user for something in text

    Parameters
    ----------
    prompt : str
        Prompt message
    regex : str
        Regular expression for validate user input
    on_error : str
        Message for invalid input
    ex_verifier : Callable[[str], bool] | None
        Extra verifier in addition to regex
    max_attempts : int | None
        Max count for attempts. None to infinite

    Returns
    -------
    str
        Valid user input
    """

    attempt = 0
    while True:
        if max_attempts and attempt >= max_attempts:
            return ""
        try:
            user_input = input(prompt)
            if (re.search(regex, user_input) is None) or (
                ex_verifier is not None and not ex_verifier(user_input)
            ):
                raise ValueError(on_error)
            break
        except ValueError as e:
            print(e)
        attempt = attempt + 1
    return user_input


def get_ports() -> list[str]:
    """Get port informations

    Returns
    -------
    list[str]
        List of port names
    """

    comports: list[list_ports.ListPortInfo] = list_ports.comports()
    return [port.device for port in comports]


def is_there_some_ports() -> bool:
    """Check if there is some ports exists

    Returns
    -------
    bool
        True if exists
    """

    return bool(len(get_ports()))


def ask_user_for_port() -> str:
    """Ask the user for the port to use

    If there's only one port, auto selects without asking.

    Returns
    -------
    str
        Port name (COM or fd)

    Notes
    -----
        If the console is not available, raise EnvironmentError.
        If there's no ports, raise IOError.
    """
    if not sys.stdin.isatty():
        raise EnvironmentError("There's no console.")

    ports: list[list_ports.ListPortInfo] = list_ports.comports()

    if ports == []:
        raise IOError("There's no serial port.")

    if len(ports) == 1:
        print(f"Auto selected: {ports[0].device}")
        return str(ports[0].device)

    print("Multiple ports detected.")
    for i, port in enumerate(ports):
        if port.manufacturer == "MONOWIRELESS":
            # Print in magenta
            print(f"[{i+1}] {port.device} \033[35m{port.description} (Genuine)\033[00m")
        elif port.manufacturer == "TOCOS":
            # Print in blue
            print(f"[{i+1}] {port.device} \033[34m{port.description} (Legacy)\033[00m")
        else:
            print(f"[{i+1}] {port.device} {port.description}")

    user_input = ask_user(
        f"Select [1-{len(ports)}]: ",
        regex=r"^[0-9]+$",
        ex_verifier=lambda s: int(s) <= len(ports),
        on_error=f"Please answer 1-{len(ports)}.",
    )
    print(f"Selected: {ports[int(user_input) - 1].device}")
    return str(ports[int(user_input) - 1].device)


def millis() -> int:
    """Get current time in milliseconds

    Returns
    -------
    int
        Current epoch in milliseconds
    """

    return round(time.time_ns() / 1000000)


def lrc8(data: bytes) -> int:
    """Calculate 8-bit LRC for given data

    Parameters
    ----------
    data : bytes
        Bytes to calculate

    Returns
    -------
    int
        LRC checksum
    """

    return int(((sum(data) ^ 0xFF) + 1) & 0xFF)


def hex_from(character: int) -> int:
    """Convert to hex from character

    Parameters
    ----------
    character : int
        Integer value of an ASCII character ('0'-'F')

    Returns
    -------
    int
        Binary integer value (0x0-0xF)
    """

    return character - ord("0") if character < ord("A") else character - ord("A") + 0xA


def character_from(hexvalue: int) -> int:
    """Convert to character from hex

    Parameters
    ----------
    hexvalue : int
        Binary integer value (0x0-0xF)

    Returns
    -------
    int
        Integer value of an ASCII character ('0'-'F')
    """

    return ord("0") + hexvalue if hexvalue < 0xA else ord("A") + hexvalue - 0xA


def byte_count_from(character_count: int) -> int:
    """Convert to byte count from character count

    Parameters
    ----------
    character_count : int
        bytes count in ascii format

    Returns
    -------
    int
        bytes count in binary format
    """

    return math.floor(character_count / 2)


def is_initialized(port: serial.Serial) -> bool:
    """Check if the serial port is initialized

    Parameters
    ----------
    port : serial.Serial
        pyserial instance

    Returns
    -------
    bool
        initialized if true
    """

    return bool(port.readable() and port.writable())


def is_readable(port: serial.Serial) -> bool:
    """Check if the serial port is readable

    Parameters
    ----------
    port : serial.Serial
        pyserial instance

    Returns
    -------
    bool
        readable if true
    """

    return bool(port.readable())


def is_writable(port: serial.Serial) -> bool:
    """Check if the serial port is writable

    Parameters
    ----------
    port : serial.Serial
        pyserial instance

    Returns
    -------
    bool
        writable if true
    """

    return bool(port.writable())


def write_binary(port: serial.Serial, data: int | bytes) -> None:
    """Write binary integer value to the serial port

    Parameters
    ----------
    port : serial.Serial
        pyserial instance
    data : int | bytes
        Binary integer value or bytes
    """

    match data:
        case int():
            if not is_writable(port):
                return None
            if not data >= 0:
                return None
            elif data <= 0xFF:
                port.write(data.to_bytes(1))
            elif data <= 0xFFFF:
                port.write(data.to_bytes(2))
            elif data <= 0xFFFFFFFF:
                port.write(data.to_bytes(4))
        case bytes():
            for byte in data:
                write_binary(port, byte)


def write_in_ascii(port: serial.Serial, data: int | bytes) -> None:
    """Write binary in ASCII format to the serial port

    Parameters
    ----------
    port : serial.Serial
        pyserial instance
    data : int
        Binary integer value or bytes
    """

    match data:
        case int():
            if not is_writable(port):
                return None
            if not data >= 0:
                return None
            elif data <= 0xFF:
                port.write(
                    bytes(
                        [character_from((data >> i) & 0x0F) for i in range(4, -1, -4)]
                    )
                )
            elif data <= 0xFFFF:
                port.write(
                    bytes(
                        [character_from((data >> i) & 0x0F) for i in range(12, -1, -4)]
                    )
                )
            elif data <= 0xFFFFFFFF:
                port.write(
                    bytes(
                        [character_from((data >> i) & 0x0F) for i in range(28, -1, -4)]
                    )
                )
        case bytes():
            for byte in data:
                write_in_ascii(port, byte)


def flush_rx_buffer(port: serial.Serial) -> None:
    """Flush serial rx buffer

    Parameters
    ----------
    port : serial.Serial
        pyserial instance
    """

    port.reset_input_buffer()


def flush_tx_buffer(port: serial.Serial) -> None:
    """Flush serial tx buffer

    Parameters
    ----------
    port : serial.Serial
        pyserial instance
    """

    port.reset_output_buffer()


def find_binary(
    port: serial.Serial,
    data: bytes,
    timeout: int,
    with_terminal: bool = False,
    terminal: int = 0,
    debugging: bool = False,
) -> bool:
    """Find binary bytes in serial rx data

    Parameters
    ----------
    port : serial.Serial
        pyserial instance
    data : bytes
        Binary data bytes to find
    timeout : int
        Timeout in seconds
    with_terminal : bool
        Use terminal byte for data input or not
    terminal : int
        Terminal byte for data input
    debugging : bool
        Print debug info if true

    Returns
    -------
    bool
        Found if true
    """

    if not is_initialized(port):
        return False
    if not len(data) > 0:
        return False
    timestamp: int = millis()
    while True:
        if millis() - timestamp > timeout * 1000:
            return False
        if port.in_waiting:
            read_byte = port.read()
            if debugging:
                print(read_byte.decode("utf-8"))
            if int.from_bytes(read_byte) == data[0]:  # compare as int
                break
    for datum in data[1:]:  # omit the first one
        if with_terminal and terminal == datum:
            break
        while True:
            if millis() - timestamp > timeout * 1000:
                return False
            if port.in_waiting:
                read_byte = port.read()
                if debugging:
                    print(read_byte.decode("utf-8"))
                if int.from_bytes(read_byte) == datum:
                    break
                else:
                    # If data is invalid, retry recursively.
                    return find_binary(
                        port,
                        data,
                        timeout * 1000 - millis() + timestamp,
                        with_terminal,
                        terminal,
                        debugging,
                    )
    return True


def find_ascii(
    port: serial.Serial, data: str, timeout: int, debugging: bool = False
) -> bool:
    """Find ASCII-formatted bytes in serial rx data

    Parameters
    ----------
    port : serial.Serial
        pyserial instance
    data : str
        ASCII-formatted bytes to find
    timeout : int
        Timeout in seconds
    debugging : bool
        Print debug info if true

    Returns
    -------
    bool
        Found if true
    """

    return find_binary(port, data.encode("ascii"), timeout, False, 0, debugging)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# MWings parsers

# Redundant imports for mypy
from . import act as act
from . import app_twelite as app_twelite
from . import app_io as app_io
from . import app_aria as app_aria
from . import app_cue as app_cue
from . import app_cue_pal_event as app_cue_pal_event
from . import app_pal_openclose as app_pal_openclose
from . import app_pal_amb as app_pal_amb
from . import app_pal_mot as app_pal_mot
from . import app_uart_ascii as app_uart_ascii
from . import app_uart_ascii_extended as app_uart_ascii_extended



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet parser for App_Twelite

from datetime import datetime
from typing import Any, final

from overrides import override
from pydantic import Field, field_validator

from .. import common


@final
class ParsedPacket(common.ParsedPacketBase):
    """Dataclass for parsed packets from App_Twelite

    Attributes
    ----------
    destination_logical_id: common.UInt8
        Logical ID for the destination (parent) device
    relay_count: common.UInt8
        Number of relay stations
    periodic: bool
        True if the packet is a periodic transmit packet
    di_changed: common.CrossSectional[bool]
        State for each digital interfaces; True if changed
    di_state: common.CrossSectional[bool]
        Input status for each digital interfaces
    ai_voltage: common.CrossSectional[common.UInt16]
        Voltage in mV for each analog interfaces
    """

    destination_logical_id: common.UInt8 = Field(default=common.UInt8(0x78))
    relay_count: common.UInt8 = Field(
        default=common.UInt8(0), ge=common.UInt8(0), le=common.UInt8(3)
    )
    periodic: bool = Field(default=False)
    di_changed: common.CrossSectional[bool] = Field(
        default=common.CrossSectional[bool](4, [False for _ in range(4)])
    )
    di_state: common.CrossSectional[bool] = Field(
        default=common.CrossSectional[bool](4, [False for _ in range(4)])
    )
    ai_voltage: common.CrossSectional[common.UInt16] = Field(
        default=common.CrossSectional[common.UInt16](
            4, [common.UInt16(0) for _ in range(4)]
        )
    )

    @field_validator("destination_logical_id")
    @classmethod
    def check_destination_logical_id(cls, lid: int) -> int:
        """Check destination lid

        Must be in range between 0 and 100 or 120 and 127 but 122
        (Router) is invalid

        Parameters
        ----------
        lid : int
            Input

        Returns
        -------
        int
            Valid input

        Raises
        ------
        ValueError
            Out of range
        """

        if 0 <= lid <= 100:
            return lid
        elif lid in range(120, 128) and lid != 122:
            return lid
        else:
            raise ValueError("must be in range (0-100) or (120-121/123-127)")

    @field_validator("ai_voltage")
    @classmethod
    def check_ai_voltage(
        cls, aiv: common.CrossSectional[int]
    ) -> common.CrossSectional[int]:
        """Check voltage of analog interfaces

        Must be in range between 0 and 3700 (VCCmax3600+margin100)

        Parameters
        ----------
        aiv : common.CrossSectional
            Input

        Returns
        -------
        common.CrossSectional
            Valid input

        Raises
        ------
        ValueError
            Out of range
        """

        for voltage in aiv:
            if voltage < 0 or voltage > 3700:
                raise ValueError("Out of range")
        return aiv


@final
class PacketParser(common.PacketParserBase):
    """Packet parser for App_Twelite"""

    @staticmethod
    @override
    def is_valid(bare_packet: common.BarePacket) -> bool:
        """Check the given bare packet is valid or not

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Static overridden method
        """
        if (
            bare_packet.u8_at(1) == 0x81
            and bare_packet.u8_at(3) == 0x01
            and (bare_packet.u8_at(5) & 0x80) == 0x80
            and len(bare_packet.payload) == 23
        ):
            return True
        return False

    @staticmethod
    @override
    def parse(bare_packet: common.BarePacket) -> ParsedPacket | None:
        """Try to parse the given bare packet

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        ParsedPacket | None
            Parsed packet data if valid else None

        Notes
        -----
        Static overridden method
        """
        if not PacketParser.is_valid(bare_packet):
            return None
        parsed_packet_data: dict[str, Any] = {
            "time_parsed": datetime.now(common.Timezone),
            "packet_type": common.PacketType.APP_TWELITE,
            "sequence_number": bare_packet.u16_at(10),  # In this, timestamp
            "source_serial_id": bare_packet.u32_at(5),
            "source_logical_id": bare_packet.u8_at(0),
            "lqi": bare_packet.u8_at(4),
            "supply_voltage": bare_packet.u16_at(13),
            "destination_logical_id": bare_packet.u8_at(9),
            "relay_count": bare_packet.u8_at(12),
            "periodic": ((bare_packet.u8_at(16) & 0x80) == 0x80),
            "di_state": common.CrossSectional[bool](
                4,
                [bool(bare_packet.u8_at(16) & (1 << port)) for port in range(4)],
            ),
            "di_changed": common.CrossSectional[bool](
                4,
                [bool(bare_packet.u8_at(17) & (1 << port)) for port in range(4)],
            ),
            "ai_voltage": common.CrossSectional[common.UInt16](
                4,
                [
                    common.UInt16(
                        min(
                            bare_packet.u8_at(18 + port) * 16
                            + ((bare_packet.u8_at(22) >> (port * 2)) & 0x03) * 4,
                            2000,
                        )
                    )
                    for port in range(4)
                ],
            ),
        }
        return ParsedPacket(**parsed_packet_data)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet parser for App_Uart (Mode A)

from datetime import datetime
from base64 import b64encode
from typing import Any, final

from pydantic import Field, computed_field, field_validator
from overrides import override

from .. import common


@final
class ParsedPacket(common.ParsedPacketBase):
    """Dataclass for parsed packets from App_Uart (Mode A)

    Attributes
    ----------
    command_id: common.UInt8
        Command ID
    data: bytes
        Data body (Hidden in JSON or something)
    data_base64: str
        Data body in Base64 for JSON or something
    data_hexstr: str
        Data body in ASCII string
    """

    command_id: common.UInt8 = Field(
        default=common.UInt8(0x00), ge=common.UInt8(0x00), lt=common.UInt8(0x80)
    )
    data: bytes = Field(default=bytes(), exclude=True)

    @computed_field
    def data_base64(self) -> str:
        return b64encode(self.data).decode("ascii")

    @computed_field
    def data_hexstr(self) -> str:
        return self.data.hex().upper()

    @field_validator("data")
    @classmethod
    def check_data(cls, data: bytes) -> bytes:
        """Check for data

        Parameters
        ----------
        data : bytes
            Input

        Returns
        -------
        bytes
            Valid input

        Raises
        ------
        ValueError
            Byte length is not in range between 1 and 80
        """

        if not (1 <= len(data) <= 80):
            raise ValueError("Payload length should be in range between 1 and 80.")
        return data


@final
class PacketParser(common.PacketParserBase):
    """Packet parser for App_Uart (Mode A)"""

    @staticmethod
    @override
    def is_valid(bare_packet: common.BarePacket) -> bool:
        """Check the given bare packet is valid or not

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Static overridden method
        """
        if (
            (0x00 <= bare_packet.u8_at(0) <= 0x64 or bare_packet.u8_at(0) == 0x78)
            and bare_packet.u8_at(1) < 0x80
            and 3 <= len(bare_packet.payload) <= 82
        ):
            return True
        return False

    @staticmethod
    @override
    def parse(bare_packet: common.BarePacket) -> ParsedPacket | None:
        """Try to parse the given bare packet

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        ParsedPacket | None
            Parsed packet data if valid else None

        Notes
        -----
        Static overridden method
        """
        if not PacketParser.is_valid(bare_packet):
            return None
        parsed_packet_data: dict[str, Any] = {
            "time_parsed": datetime.now(common.Timezone),
            "packet_type": common.PacketType.APP_UART_ASCII,
            "sequence_number": None,
            "source_serial_id": common.UInt32(0),
            "source_logical_id": bare_packet.u8_at(0),
            "lqi": None,
            "supply_voltage": None,
            "command_id": bare_packet.u8_at(1),
            "data": bare_packet.payload[2:],
        }
        return ParsedPacket(**parsed_packet_data)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# MWings

from enum import IntEnum, auto
from threading import Thread, Event
from datetime import timezone, tzinfo
from typing import Any, Callable, Self, overload, final
from types import TracebackType
from warnings import warn

import serial  # type: ignore
from pyee.base import EventEmitter
from overrides import override

from . import common
from . import utils
from . import parsers
from . import serializers


@final
class Twelite(Thread):
    """MWings main class"""

    # Private inner enum classes
    @final
    class __State(IntEnum):
        """Parser state

        Attributes
        ----------
        WAITING_FOR_HEADER: int
            Waiting for the header character ':'
        RETRIEVING_PAYLOAD: int
            Retrieving packet payload data
        WAITING_FOR_FOOTER: int
            Waiting for the footer character <LF>
        COMPLETED: int
            Completed to receive a packet
        CHECKSUM_ERROR: int
            Error in checksum
        UNKNOWN_ERROR: int
            Unknown error
        """

        WAITING_FOR_HEADER = auto()
        RETRIEVING_PAYLOAD = auto()
        WAITING_FOR_FOOTER = auto()
        COMPLETED = auto()
        UNKNOWN_ERROR = auto()
        CHECKSUM_ERROR = auto()
        TIMEOUT_ERROR = auto()

    @final
    class __Command(IntEnum):
        """Revised 0xDB commands

        Attributes
        ----------
        ACK: int
            Revised 0xDB command: Ack
        MODULE_ADDRESS: int
            Revised 0xDB command: Module address
        SET_PARAMETER: int
            Revised 0xDB command: Set parameter
        GET_PARAMETER: int
            Revised 0xDB command: Get parameter
        CONTROL: int
            Revised 0xDB command: Control module
        DISABLE_SILENT_MODE: int
            Revised 0xDB command: Disable silent mode
        CLEAR: int
            Revised 0xDB command: Clear settings
        SAVE: int
            Revised 0xDB command: Save settings
        RESET: int
            Revised 0xDB command: Reset module
        """

        ACK = 0xD0
        MODULE_ADDRESS = 0xD1
        SET_PARAMETER = 0xD2
        GET_PARAMETER = 0xD3
        CONTROL = 0xD8
        DISABLE_SILENT_MODE = 0xD9
        CLEAR = 0xDD
        SAVE = 0xDE
        RESET = 0xDF

    @final
    class __Parameter(IntEnum):
        """Parameter for commands: set/get parameter

        Attributes
        ---------
        APP_ID: int
            Application ID
        CH_MASK: int
            Channels bit mask
        RETRY_TX: int
            Retry count / Tx power
        ROUTING_LAYER: int
            Routing layer
        AP_ADDRESS: int
            Access point address
        UART_BAUDRATE: int
            Uart baudrate
        ENCRYPTION: int
            Encryption settings
        OPTION_BITS: int
            Option bits
        """

        APP_ID = 0x01
        CH_MASK = 0x02
        RETRY_TX = 0x03
        ROUTING_LAYER = 0x04
        AP_ADDRESS = 0x05
        UART_BAUDRATE = 0x06
        ENCRYPTION = 0x07
        OPTION_BITS = 0x08

    @final
    class __SavingStatus(IntEnum):
        """Return status during saving parameters

        Attributes
        ----------
        SUCCEEDED: int
            Succeeded
        FAILED: int
            Failed
        SUCCEEDED_NO_MODIFICATIONS: int
            Succeeded, but no modifications
        FAILED_NO_MODIFICATIONS: int
            Failed, but no modifications
        """

        SUCCEEDED = 0x01
        FAILED = 0x00
        SUCCEEDED_NO_MODIFICATIONS = 0x81
        FAILED_NO_MODIFICATIONS = 0x80

    # Private varibles

    # pyserial instance
    __serial: Any = None

    # Rx binary data buffer
    __buffer: bytearray

    # Max size of the rx buffer
    __rx_buffer_size: int

    # Number of received ASCII characters
    __character_count: int

    # LRC checksum for the latest received packet
    __checksum: int

    # Timeout for each packet in milliseconds
    __timeout: int

    # Timestamp of the last time the header was received
    __latest_timestamp: int

    # Print debug info if True
    __debugging: bool

    # Parser state
    __state: __State

    # Event emitter for receiving events
    __event_emitter: EventEmitter

    # Status for the receiver thread
    __running: bool

    # Set to ensure the receiver thread has stopped
    __ensure_stopped: Event

    # Public methods

    def __init__(
        self,
        port: str | None = None,
        rx_buffer_size: int = 1024,
        timeout: int = 100,
        tz: tzinfo | None = None,
        debugging: bool = False,
    ):
        """Constructor

        Parameters
        ----------
        port : str | None
            Name for the serial port to use / set None to disable serial
        rx_buffer_size : int
            Receive buffer size
        timeout : int
            Timeout for each packet in milliseconds
        tz : tzinfo
            Timezone for datetime data. Default is UTC (Aware). Use ZoneInfo() for others.
        debugging : bool
            Print debug info if true
        """

        super().__init__()
        self.daemon = False

        if port is not None:
            try:
                self.__serial = serial.Serial(port, 115200, timeout=1)
            except serial.serialutil.SerialException:
                raise IOError("Specified port is busy or not available")

        self.__buffer = bytearray()
        self.__rx_buffer_size = rx_buffer_size
        self.__character_count = 0
        self.__checksum = 0
        self.__timeout = timeout
        self.__latest_timestamp = -1
        self.__debugging = debugging
        self.__state = self.__State.WAITING_FOR_HEADER
        self.__event_emitter = EventEmitter()
        self.__running = False
        self.__ensure_stopped = Event()
        common.Timezone = tz if tz is not None else timezone.utc

    def __del__(self) -> None:
        """Destructor"""
        self.close()

    def __enter__(self) -> Self:
        """Enter the runtime context related to this object."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        """Exit the runtime context and close the connection."""
        self.close()
        return None

    def close(self) -> None:
        """Close the serial port if opened"""
        if self.__serial is not None:
            if self.__running:
                self.stop()
            if self.__serial.is_open:
                self.__serial.close()

    def clear_input(self) -> None:
        """Clear input buffer of serial and packets"""
        if self.__serial is not None and self.__serial.is_open:
            self.__serial.reset_input_buffer()
        self.__buffer.clear()

    @staticmethod
    def set_timezone(tz: tzinfo | None) -> None:
        """Set timezone for received data

        Parameters
        ----------
        tz : tzinfo
            tzinfo object. Typically Zoneinfo("IANA/City"). None for UTC.
        """

        common.Timezone = tz if tz is not None else timezone.utc

    @property
    def timezone(self) -> tzinfo:
        """Get timezone set

        Returns
        -------
        tzinfo
            Timezone set for mwings
        """

        return common.Timezone

    @overload
    def add_listener(
        self, event: common.PacketType, handler: Callable[[common.BarePacket], None]
    ) -> None:
        """Register a handler for receiving packets

        Parameters
        ----------
        event : common.PacketType
            Identifier for packets to receive
        handler : Callable[[common.BarePacket], None]
            Handler to handle bare packets
        """
        ...

    @overload
    def add_listener(
        self,
        event: common.PacketType,
        handler: Callable[[common.SomeParsedPacket], None],
    ) -> None:
        """Register a handler for receiving packets

        Parameters
        ----------
        event : common.PacketType
            Identifier for packets to receive
        handler : Callable[[common.SomeParsedPacket], None]
            Handler to handle some parsed packets
        """
        ...

    def add_listener(
        self, event: common.PacketType, handler: common.SomeCallable
    ) -> None:
        """Register a handler for receiving packets

        Parameters
        ----------
        event : common.PacketType
            Identifier for packets to receive
        handler : common.SomeCallable
            Handler to handle packets
        """

        self.__event_emitter.add_listener(event, handler)

    def remove_all_listeners(self, event: common.PacketType | None) -> None:
        """Remove all handlers for receicing packets

        Parameters
        ----------
        event : common.PacketType
            Identifier for packets to receive
        """

        self.__event_emitter.remove_all_listeners(event)

    def on(
        self, event: common.PacketType
    ) -> Callable[[common.SomeCallable], common.SomeCallable]:
        """Generate a decorator to register a handler for receiving packets

        Parameters
        ----------
        event : common.PacketType
            Identifier for packets to receive

        Returns
        -------
        Callable[[common.SomeCallable], common.SomeCallable]
            Decorator to register a handler for receiving packets
        """

        def decorator(handler: common.SomeCallable) -> common.SomeCallable:
            """Decorator to register a handler for receiving packets

            Parameters
            ----------
            handler : common.SomeCallable
                Original handler for receiving packets

            Returns
            -------
            common.SomeCallable
                Decorated handler for receiving packets (Actually, it's same)
            """

            self.add_listener(event, handler)
            return handler

        return decorator

    @overload
    def send(self, data: common.BarePacket) -> bool:
        """Send data to the device with ModBus format

        Parameters
        ----------
        data : common.BarePacket
            Payload and checksum data to send

        Returns
        -------
        bool
            True if succeeded
        """
        ...

    @overload
    def send(self, data: common.SomeCommand) -> bool:
        """Send data to the device with ModBus format

        Parameters
        ----------
        data : common.SomeCommand
            Some command to serialize and send

        Returns
        -------
        bool
            True if succeeded
        """
        ...

    def send(self, data: Any) -> bool:
        """Send data to the device with ModBus format

        Parameters
        ----------
        data : Any
            Data to send

        Returns
        -------
        bool
            True if succeeded
        """

        if self.__serial is None:
            raise RuntimeError("send() can only be used when the port is initialized")

        if not utils.is_writable(self.__serial):
            return False

        match data:
            case common.BarePacket():
                utils.write_binary(self.__serial, ord(":"))
                utils.write_in_ascii(self.__serial, data.payload)
                utils.write_in_ascii(self.__serial, data.checksum)
                utils.write_binary(self.__serial, ord("\r"))
                utils.write_binary(self.__serial, ord("\n"))
                if self.__debugging:
                    print(f"Sent ascii: {data.payload.hex()}")
            case serializers.app_twelite.Command():
                serialized_data = serializers.app_twelite.CommandSerializer.serialize(
                    data
                )
                if serialized_data is None:
                    return False
                else:
                    self.send(serialized_data)
            case serializers.app_io.Command():
                serialized_data = serializers.app_io.CommandSerializer.serialize(data)
                if serialized_data is None:
                    return False
                else:
                    self.send(serialized_data)
            case serializers.app_pal_notice.Command():
                serialized_data = (
                    serializers.app_pal_notice.CommandSerializer.serialize(data)
                )
                if serialized_data is None:
                    return False
                else:
                    self.send(serialized_data)
            case serializers.app_pal_notice_detailed.Command():
                serialized_data = (
                    serializers.app_pal_notice_detailed.CommandSerializer.serialize(
                        data
                    )
                )
                if serialized_data is None:
                    return False
                else:
                    self.send(serialized_data)
            case serializers.app_pal_notice_event.Command():
                serialized_data = (
                    serializers.app_pal_notice_event.CommandSerializer.serialize(data)
                )
                if serialized_data is None:
                    return False
                else:
                    self.send(serialized_data)
            case serializers.app_uart_ascii.Command():
                serialized_data = (
                    serializers.app_uart_ascii.CommandSerializer.serialize(data)
                )
                if serialized_data is None:
                    return False
                else:
                    self.send(serialized_data)
        return True

    @override
    def start(self) -> None:
        """Start the thread to receive continuously

        Notes
        -----
        Overrides threading.Thread.start()
        """

        if self.__serial is None:
            raise RuntimeError("start() can only be used when the port is initialized")

        self.__running = True
        super().start()

    @override
    def run(self) -> None:
        """Run the thread to receive continuously

        Call this function via Twelite.start()

        Notes
        -----
        Overrides threading.Thread.run()
        """

        while self.__running:
            self.update()
        self.__ensure_stopped.set()

    def stop(self) -> None:
        """Stop the thread to receive continuously"""

        if self.__serial is None:
            raise RuntimeError("stop() can only be used when the port is initialized")

        self.__ensure_stopped.clear()
        self.__running = False
        self.__ensure_stopped.wait()

    def receive(self) -> common.PacketType:
        """Wait for parsing of a single packet

        Returns
        -------
        common.PacketType
            Identifier for packet received

        Notes
        -----
        This function blocks current thread
        """

        if self.__serial is None:
            raise RuntimeError(
                "receive() can only be used when the port is initialized"
            )

        while True:
            if (packet_type := self.update()) is not None:
                return packet_type

    def update(self) -> common.PacketType | None:
        """Update parsing state with serial data

        Returns
        -------
        common.PacketType | None
            Returns packet type identifier if available else None
        """

        if self.__serial is None:
            raise RuntimeError("update() can only be used when the port is initialized")

        # Abort if the serial is not initialized
        if not utils.is_initialized(self.__serial):
            return None

        # Process all byte in the buffer
        while utils.is_readable(self.__serial):
            # Parse a read character
            result = self.__parse(int.from_bytes(self.__serial.read()))
            if result is not None:
                # When parsed a complete packet
                return result
        return None

    def parse_line(self, line: str, use_lf: bool = True) -> common.PacketType | None:
        """Parse a single string

        Useful when use local log file instead of serial port rx

        Parameters
        ----------
        line : str
            String line
        use_lf : bool
            Use LF instead of CRLF (For f.readline, set True as
            default)

        Returns
        -------
        common.PacketType | None
            Latest packet type identifier if available else None

        Raises
        ------
        RuntimeError
            Serial IS initialized
        """
        if self.__serial is not None:
            raise RuntimeError(
                "parse_line() can only be used when the port is NOT initialized"
            )
        result: common.PacketType | None = None
        for character in line.lstrip():
            # Parse a read character
            latest_result = self.__parse(character, use_lf)
            if latest_result is not None:
                # When parsed a complete packet
                result = latest_result
        return result

    @overload
    def parse(self, character: str, use_lf: bool = False) -> common.PacketType | None:
        ...

    @overload
    def parse(self, character: bytes, use_lf: bool = False) -> common.PacketType | None:
        ...

    @overload
    def parse(self, character: int, use_lf: bool = False) -> common.PacketType | None:
        ...

    def parse(self, character: Any, use_lf: bool = False) -> common.PacketType | None:
        """Parse a single character

        Parameters
        ----------
        character : Any
            character to parse
        use_lf : bool
            If use LF instead of CRLF, set True

        Returns
        -------
        common.PacketType | None
            If completed, returns packet type identifier

        Raises
        ------
        RuntimeError
            Serial IS initialized
        ValueError
            Unsupported character
        """
        if self.__serial is not None:
            raise RuntimeError(
                "parse() can only be used when the port is NOT initialized"
            )
        return self.__parse(character, use_lf)

    # Private method(s)

    @overload
    def __parse(self, character: str, use_lf: bool = False) -> common.PacketType | None:
        ...

    @overload
    def __parse(
        self, character: bytes, use_lf: bool = False
    ) -> common.PacketType | None:
        ...

    @overload
    def __parse(self, character: int, use_lf: bool = False) -> common.PacketType | None:
        ...

    def __parse(self, character: Any, use_lf: bool = False) -> common.PacketType | None:
        """Parse a single character

        Parameters
        ----------
        character : Any
            character to parse
        use_lf : bool
            If use LF instead of CRLF, set True

        Returns
        -------
        common.PacketType | None
            If completed, returns packet type identifier

        Raises
        ------
        ValueError
            Unsupported character
        """
        match character:
            case str():
                if len(character) != 1:
                    raise ValueError("character must be single length")
                character_bytes: bytes
                try:
                    character_bytes = character.encode("ascii")
                except UnicodeEncodeError:
                    raise ValueError("character must bpe ASCII")
                return self.__parse(character_bytes, use_lf)
            case bytes():
                if len(character) != 1:
                    raise ValueError("character must be single length")
                return self.__parse(character[0], use_lf)
            case int():
                # Abort if the read byte is invalid
                if not (0 <= character <= 0xFF):
                    return None
                if self.__debugging:
                    pass
                    # print(repr(chr(character)))

                # Process packet contents upon parsing completion
                if (bare_packet := self.__process_ascii(character, use_lf)) is not None:
                    # Emit events here

                    # Bare packet handler
                    self.__event_emitter.emit(common.PacketType.BARE, bare_packet)

                    # Act packet handler
                    if parsers.act.PacketParser.is_valid(bare_packet):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.ACT,
                                parsers.act.PacketParser.parse(bare_packet),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.ACT}"
                            )
                        return common.PacketType.ACT

                    # App_Twelite packet handler
                    if parsers.app_twelite.PacketParser.is_valid(bare_packet):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.APP_TWELITE,
                                parsers.app_twelite.PacketParser.parse(bare_packet),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.APP_TWELITE}"
                            )
                        return common.PacketType.APP_TWELITE

                    # App_Io packet handler
                    if parsers.app_io.PacketParser.is_valid(bare_packet):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.APP_IO,
                                parsers.app_io.PacketParser.parse(bare_packet),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.APP_IO}"
                            )
                        return common.PacketType.APP_IO

                    # App_ARIA packet handler
                    if parsers.app_aria.PacketParser.is_valid(bare_packet):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.APP_ARIA,
                                parsers.app_aria.PacketParser.parse(bare_packet),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.APP_ARIA}"
                            )
                        return common.PacketType.APP_ARIA

                    # App_CUE packet handler
                    if parsers.app_cue.PacketParser.is_valid(bare_packet):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.APP_CUE,
                                parsers.app_cue.PacketParser.parse(bare_packet),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.APP_CUE}"
                            )
                        return common.PacketType.APP_CUE

                    # App_CUE (PAL Move or Dice mode) packet handler
                    if parsers.app_cue_pal_event.PacketParser.is_valid(bare_packet):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.APP_CUE_PAL_EVENT,
                                parsers.app_cue_pal_event.PacketParser.parse(
                                    bare_packet
                                ),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.APP_CUE_PAL_EVENT}"
                            )
                        return common.PacketType.APP_CUE_PAL_EVENT

                    # App_PAL (OPENCLOSE) packet handler
                    if parsers.app_pal_openclose.PacketParser.is_valid(bare_packet):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.APP_PAL_OPENCLOSE,
                                parsers.app_pal_openclose.PacketParser.parse(
                                    bare_packet
                                ),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.APP_PAL_OPENCLOSE}"
                            )
                        return common.PacketType.APP_PAL_OPENCLOSE

                    # App_PAL (AMB) packet handler
                    if parsers.app_pal_amb.PacketParser.is_valid(bare_packet):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.APP_PAL_AMB,
                                parsers.app_pal_amb.PacketParser.parse(bare_packet),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.APP_PAL_AMB}"
                            )
                        return common.PacketType.APP_PAL_AMB

                    # App_PAL (MOT) packet handler
                    if parsers.app_pal_mot.PacketParser.is_valid(bare_packet):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.APP_PAL_MOT,
                                parsers.app_pal_mot.PacketParser.parse(bare_packet),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.APP_PAL_MOT}"
                            )
                        return common.PacketType.APP_PAL_MOT

                    # App_Uart (Mode A) packet handler
                    if parsers.app_uart_ascii.PacketParser.is_valid(bare_packet):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.APP_UART_ASCII,
                                parsers.app_uart_ascii.PacketParser.parse(bare_packet),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.APP_UART_ASCII}"
                            )
                        return common.PacketType.APP_UART_ASCII

                    # App_Uart (Mode A, Extended) packet handler
                    if parsers.app_uart_ascii_extended.PacketParser.is_valid(
                        bare_packet
                    ):
                        if (
                            self.__event_emitter.emit(
                                common.PacketType.APP_UART_ASCII_EXTENDED,
                                parsers.app_uart_ascii_extended.PacketParser.parse(
                                    bare_packet
                                ),
                            )
                            is not True
                            and self.__debugging
                        ):
                            warn(
                                f"No handler(s) registered for {common.PacketType.APP_UART_ASCII_EXTENDED}"
                            )
                        return common.PacketType.APP_UART_ASCII_EXTENDED

                    return common.PacketType.BARE
        return None

    def __process_ascii(
        self, character: int, use_lf: bool = False
    ) -> common.BarePacket | None:
        """Process an ascii character

        Parameters
        ----------
        character : int
            ASCII code for the received character
        use_lf : bool
            If use LF instead of CRLF, set True

        Returns
        -------
        common.BarePacket | None
            Complete bare packet if received the whole packet else None
        """

        # Reset if the state is error or completed
        if (
            self.__state == self.__State.COMPLETED
            or self.__state == self.__State.UNKNOWN_ERROR
            or self.__state == self.__State.CHECKSUM_ERROR
            or self.__state == self.__State.TIMEOUT_ERROR
        ):
            self.__state = self.__State.WAITING_FOR_HEADER

        # Reset on timeout
        if (
            self.__timeout > 0
            and self.__state != self.__State.WAITING_FOR_HEADER
            and utils.millis() - self.__latest_timestamp > self.__timeout
        ):
            self.__state = self.__State.TIMEOUT_ERROR
            if self.__debugging:
                print("TIMEOUT_ERROR\n")

        # Run state machine
        match self.__state:
            case self.__State.WAITING_FOR_HEADER:
                # If the character is colon, start to read
                if character == ord(":"):
                    self.__state = self.__State.RETRIEVING_PAYLOAD
                    self.__latest_timestamp = utils.millis()
                    self.__character_count = 0
                    self.__checksum = 0
                    self.__buffer = bytearray()
            case self.__State.RETRIEVING_PAYLOAD:
                if (character >= ord("0") and character <= ord("9")) or (
                    character >= ord("A") and character <= ord("F")
                ):
                    # Valid hex character

                    # Abort if the buffer is overflowing
                    if (
                        utils.byte_count_from(self.__character_count)
                        >= self.__rx_buffer_size
                    ):
                        self.__state = self.__State.UNKNOWN_ERROR
                        if self.__debugging:
                            print("OVERFLOW ERROR\n")

                    # Convert character to hex
                    hex_value: int = utils.hex_from(character)

                    # Get an index for the new byte
                    newByteIndex: int = (
                        utils.byte_count_from(self.__character_count) + 1 - 1
                    )  # next position, but zero origin

                    # Add byte
                    if self.__character_count & 1:
                        # Odd: set 0-3 bit of the new byte
                        self.__buffer[newByteIndex] = (
                            self.__buffer[newByteIndex] & 0xF0
                        ) | hex_value
                        self.__checksum += self.__buffer[newByteIndex]
                    else:
                        # Even: set 7-4 bit of the byte
                        self.__buffer.append(hex_value << 4)
                    self.__character_count += 1

                elif (not use_lf and character == ord("\r")) or (
                    use_lf and character == ord("\n")
                ):
                    # Abort if received data are not valid
                    if (
                        not self.__character_count >= 4
                        and (self.__character_count & 1) == 0
                    ):
                        self.__state = self.__State.UNKNOWN_ERROR
                        if self.__debugging:
                            print("LENGTH ERROR")

                    # Mask checksum
                    self.__checksum = self.__checksum & 0xFF

                    # Abort if the checksum is not valid
                    if not self.__checksum == 0:
                        self.__state = self.__State.CHECKSUM_ERROR
                        if self.__debugging:
                            print("CHECKSUM_ERROR")

                    if not use_lf:
                        self.__state = self.__State.WAITING_FOR_FOOTER
                    else:
                        self.__state = self.__State.COMPLETED
                else:
                    # Invalid characters
                    self.__state = self.__State.UNKNOWN_ERROR
                    if self.__debugging:
                        print(f"INVALID CHAR ERROR: {repr(chr(character))}")

            case self.__State.WAITING_FOR_FOOTER:
                if character == ord("\n"):
                    # Completed
                    self.__state = self.__State.COMPLETED
                else:
                    # CR only
                    self.__state = self.__State.UNKNOWN_ERROR
                    if self.__debugging:
                        print("NO LF ERROR")
            case _:
                self.__state = self.__State.UNKNOWN_ERROR
                if self.__debugging:
                    print("UNKNOWN ERROR")

        # Make bare packet available when parsing was completed
        if self.__state == self.__State.COMPLETED:
            bare_packet_data: dict[str, Any] = {
                "payload": bytes(self.__buffer[:-1]),  # -1 for checksum
                "checksum": self.__buffer[
                    utils.byte_count_from(self.__character_count) - 1
                ],
            }
            return common.BarePacket(**bare_packet_data)
        return None



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet serializer for App_PAL (NOTICE) detailed commands

from typing import Any, final

from pydantic import Field
from overrides import override

from .. import common
from .. import utils


@final
class Command(common.CommandBase):
    """Dataclass for App_PAL (NOTICE) detailed command

    Attributes
    ----------
    color : common.AppPalNoticeRGBWColor
        Color in RGBW (0-0xF)
    blink_duty_percentage : common.UInt8
        Blink duty in %
    blink_period_in_sec : common.Float64
        Blink period in sec (0-10.2)
    duration_in_sec : common.UInt8
        Duration in seconds
    """

    color: common.AppPalNoticeRGBWColor = Field(default=common.AppPalNoticeRGBWColor())
    blink_duty_percentage: common.UInt8 = Field(
        default=common.UInt8(100), ge=common.UInt8(0), le=common.UInt8(100)
    )
    blink_period_in_sec: common.Float64 = Field(
        default=common.Float64(1.0), ge=common.Float64(0.0), le=common.Float64(10.2)
    )
    duration_in_sec: common.UInt8 = Field(
        default=common.UInt8(1), le=common.UInt8(0xFF)
    )

    @override
    def is_valid(self) -> bool:
        """Check if the command content is valid or not

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Overridden
        """

        return 0x00 <= self.destination_logical_id <= 0x64


@final
class CommandSerializer(common.CommandSerializerBase):
    """Command serializer for App_PAL (NOTICE)"""

    @staticmethod
    @override
    def serialize(command: common.SomeCommand) -> common.BarePacket | None:
        """Serialize the given command

        Parameters
        ----------
        command : common.SomeCommand
            App_PAL (NOTICE) command to serialize

        Returns
        -------
        common.BarePacket | None
            Serialized bytes and its LRC checksum (8bit) if valid

        Notes
        -----
        Static overridden method
        """

        if not (isinstance(command, Command) and command.is_valid()):
            return None

        payload_to_build: bytearray = bytearray()

        payload_to_build.append(min(command.destination_logical_id, 0x64))
        payload_to_build.append(0x90)
        payload_to_build.append(0x03)

        # Set color in RGBW
        payload_to_build.append(0x03)
        payload_to_build.append(0xFF)
        payload_to_build.append((command.color.u16() >> 8) & 0xFF)
        payload_to_build.append((command.color.u16() >> 0) & 0xFF)

        # Set blinking behavior
        payload_to_build.append(0x04)
        payload_to_build.append(0xFF)
        payload_to_build.append(round(command.blink_duty_percentage * 0xFF / 100))
        payload_to_build.append(round(command.blink_period_in_sec / 0.04))

        # Set duration
        payload_to_build.append(0x02)
        payload_to_build.append(0xFF)
        payload_to_build.append((command.duration_in_sec >> 8) & 0xFF)
        payload_to_build.append((command.duration_in_sec >> 0) & 0xFF)

        serialized_packet: dict[str, Any] = {
            "payload": bytes(payload_to_build),
            "checksum": utils.lrc8(bytes(payload_to_build)),
        }

        return common.BarePacket(**serialized_packet)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet serializer for App_PAL (NOTICE) event commands

from typing import Any, final

from pydantic import Field
from overrides import override

from .. import common
from .. import utils


@final
class Command(common.CommandBase):
    """Dataclass for App_PAL (NOTICE) event command

    Attributes
    ----------
    event_id : common.UInt8
        Event id
    """

    event_id: common.UInt8 = Field(
        default=common.UInt8(0), ge=common.UInt8(0), le=common.UInt8(0x10)
    )

    @override
    def is_valid(self) -> bool:
        """Check if the command content is valid or not

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Overridden
        """

        return 0x00 <= self.destination_logical_id <= 0x64


@final
class CommandSerializer(common.CommandSerializerBase):
    """Command serializer for App_PAL (NOTICE)"""

    @staticmethod
    @override
    def serialize(command: common.SomeCommand) -> common.BarePacket | None:
        """Serialize the given command

        Parameters
        ----------
        command : common.SomeCommand
            App_PAL (NOTICE) command to serialize

        Returns
        -------
        common.BarePacket | None
            Serialized bytes and its LRC checksum (8bit) if valid

        Notes
        -----
        Static overridden method
        """

        if not (isinstance(command, Command) and command.is_valid()):
            return None

        payload_to_build: bytearray = bytearray()

        payload_to_build.append(min(command.destination_logical_id, 0x64))
        payload_to_build.append(0x90)
        payload_to_build.append(0x01)

        # Set event id
        payload_to_build.append(0x00)
        payload_to_build.append(0x04)
        payload_to_build.append(0x00)
        payload_to_build.append(command.event_id)

        serialized_packet: dict[str, Any] = {
            "payload": bytes(payload_to_build),
            "checksum": utils.lrc8(bytes(payload_to_build)),
        }

        return common.BarePacket(**serialized_packet)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet parser for App_PAL (OPENCLOSE)

from datetime import datetime
from typing import Any, final

from overrides import override
from pydantic import Field, field_serializer

from .. import common


@final
class ParsedPacket(common.ParsedPacketBase):
    """Dataclass for parsed packets from App_PAL (OPENCLOSE)

    Attributes
    ----------
    router_serial_id: UInt32
        Serial ID for the first router device (0x80000000 with no routing)
    ai1_voltage: common.UInt16
        Voltage for AI1 port in mV
    magnet_state: common.MagnetState
        Magnet state
    magnet_state_changed: bool
        True if the magnet state was changed
    """

    router_serial_id: common.UInt32 = Field(
        default=common.UInt32(0), ge=common.UInt32(0), le=common.UInt32(0xFFFFFFFF)
    )
    ai1_voltage: common.UInt16 = Field(
        default=common.UInt16(0), ge=common.UInt16(0), le=common.UInt16(3700)
    )
    magnet_state: common.MagnetState = Field(default=common.MagnetState.NOT_DETECTED)
    magnet_state_changed: bool = Field(default=False)

    @field_serializer("magnet_state")
    def serialize_magnet_state(self, magnet_state: common.MagnetState) -> str:
        """Print magnet_state in readable names for JSON or something

        Parameters
        ----------
        magnet_state : common.MagnetState
            Magnet state

        Returns
        -------
        str
            Serialized text for JSON or something
        """

        return magnet_state.name


@final
class PacketParser(common.PacketParserBase):
    """Packet parser for App_PAL (OPENCLOSE)"""

    @staticmethod
    @override
    def is_valid(bare_packet: common.BarePacket) -> bool:
        """Check the given bare packet is valid or not

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Static overridden method
        """
        if (
            (bare_packet.u8_at(0) & 0x80) == 0x80
            and (bare_packet.u8_at(7) & 0x80) == 0x80
            and bare_packet.u8_at(12) == 0x80
            and bare_packet.u8_at(13) == 0x81
            and len(bare_packet.payload) == 33
        ):
            return True
        return False

    @staticmethod
    @override
    def parse(bare_packet: common.BarePacket) -> ParsedPacket | None:
        """Try to parse the given bare packet

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        ParsedPacket | None
            Parsed packet data if valid else None

        Notes
        -----
        Static overridden method
        """
        if not PacketParser.is_valid(bare_packet):
            return None
        parsed_packet_data: dict[str, Any] = {
            "time_parsed": datetime.now(common.Timezone),
            "packet_type": common.PacketType.APP_PAL_OPENCLOSE,
            "sequence_number": bare_packet.u16_at(5),
            "source_serial_id": bare_packet.u32_at(7),
            "source_logical_id": bare_packet.u8_at(11),
            "lqi": bare_packet.u8_at(4),
            "supply_voltage": bare_packet.u16_at(19),
            "router_serial_id": bare_packet.u32_at(0),
            "ai1_voltage": bare_packet.u16_at(25),
            "magnet_state": common.MagnetState(bare_packet.u8_at(31) & 0x0F),
            "magnet_state_changed": False if bare_packet.u8_at(31) & 0x80 else True,
        }
        return ParsedPacket(**parsed_packet_data)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet parser for App_IO

from datetime import datetime
from typing import Any, final

from overrides import override
from pydantic import Field

from .. import common


@final
class ParsedPacket(common.ParsedPacketBase):
    """Dataclass for parsed packets from App_IO

    Attributes
    ----------
    relay_count: common.UInt8
        Number of relay stations
    di_state: common.CrossSectional[bool]
        Input state for each DI ports
    di_valid: common.CrossSectional[bool]
        Valid state for each DI ports; True if used
    di_interrupt: common.CrossSectional[bool]
        Interrupt state for each DI ports; True if detected via ISR
    """

    relay_count: common.UInt8 = Field(
        default=common.UInt8(0), ge=common.UInt8(0), le=common.UInt8(3)
    )
    di_state: common.CrossSectional[bool] = Field(
        default=common.CrossSectional[bool](12, [False for _ in range(12)])
    )
    di_valid: common.CrossSectional[bool] = Field(
        default=common.CrossSectional[bool](12, [False for _ in range(12)])
    )
    di_interrupt: common.CrossSectional[bool] = Field(
        default=common.CrossSectional[bool](12, [False for _ in range(12)])
    )


@final
class PacketParser(common.PacketParserBase):
    """Packet parser for App_IO"""

    @staticmethod
    @override
    def is_valid(bare_packet: common.BarePacket) -> bool:
        """Check the given bare packet is valid or not

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Static overridden method
        """
        if (
            bare_packet.u8_at(1) == 0x81
            and bare_packet.u8_at(3) == 0x02
            and (bare_packet.u8_at(5) & 0x80) == 0x80
            and len(bare_packet.payload) == 20
        ):
            return True
        return False

    @staticmethod
    @override
    def parse(bare_packet: common.BarePacket) -> ParsedPacket | None:
        """Try to parse the given bare packet

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        ParsedPacket | None
            Parsed packet data if valid else None

        Notes
        -----
        Static overridden method
        """
        if not PacketParser.is_valid(bare_packet):
            return None
        parsed_packet_data: dict[str, Any] = {
            "time_parsed": datetime.now(common.Timezone),
            "packet_type": common.PacketType.APP_IO,
            "sequence_number": bare_packet.u16_at(10),
            "source_serial_id": bare_packet.u32_at(5),
            "source_logical_id": bare_packet.u8_at(0),
            "lqi": bare_packet.u8_at(4),
            "supply_voltage": None,  # There's no ADC
            "relay_count": bare_packet.u8_at(12),
            "di_state": common.CrossSectional[bool](
                12,
                [bool(bare_packet.u16_at(13) & (1 << port)) for port in range(12)],
            ),
            "di_valid": common.CrossSectional[bool](
                12,
                [bool(bare_packet.u16_at(15) & (1 << port)) for port in range(12)],
            ),
            "di_interrupt": common.CrossSectional[bool](
                12,
                [bool(bare_packet.u16_at(17) & (1 << port)) for port in range(12)],
            ),
        }
        return ParsedPacket(**parsed_packet_data)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet serializer for App_PAL (NOTICE) commands

from typing import Any, final

from pydantic import Field, field_serializer
from overrides import override

from .. import common
from .. import utils


@final
class Command(common.CommandBase):
    """Dataclass for App_PAL (NOTICE) command

    Attributes
    ----------
    color : common.AppPalNoticeColor
        Color
    blink_speed : common.AppPalNoticeBlinkSpeed
        Blinking speed
    brightness : common.UInt8
        Brightness from 0 to 0xF
    duration_in_sec : common.UInt8
        Duration in seconds
    """

    color: common.AppPalNoticeColor = Field(default=common.AppPalNoticeColor.WHITE)
    blink_speed: common.AppPalNoticeBlinkSpeed = Field(
        default=common.AppPalNoticeBlinkSpeed.ALWAYS_ON
    )
    brightness: common.UInt8 = Field(
        default=common.UInt8(0x8), ge=common.UInt8(0), le=common.UInt8(0xF)
    )
    duration_in_sec: common.UInt8 = Field(
        default=common.UInt8(5), ge=common.UInt8(0), le=common.UInt8(0xFF)
    )

    @field_serializer("color")
    def serialize_color(self, color: common.AppPalNoticeColor) -> str:
        """Print color in readable names for JSON or something

        Parameters
        ----------
        color : common.AppPalNoticeColor
            Color

        Returns
        -------
        str
            Serialized text for JSON or something
        """

        return color.name

    @field_serializer("blink_speed")
    def serialize_blink_speed(self, blink_speed: common.AppPalNoticeBlinkSpeed) -> str:
        """Print blink_speed in readable names for JSON or something

        Parameters
        ----------
        blink_speed : common.AppPalNoticeBlinkSpeed
            Blinking speed

        Returns
        -------
        str
            Serialized text for JSON or something
        """

        return blink_speed.name

    @override
    def is_valid(self) -> bool:
        """Check if the command content is valid or not

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Overridden
        """

        return 0x00 <= self.destination_logical_id <= 0x64


@final
class CommandSerializer(common.CommandSerializerBase):
    """Command serializer for App_PAL (NOTICE)"""

    @staticmethod
    @override
    def serialize(command: common.SomeCommand) -> common.BarePacket | None:
        """Serialize the given command

        Parameters
        ----------
        command : common.SomeCommand
            App_PAL (NOTICE) command to serialize

        Returns
        -------
        common.BarePacket | None
            Serialized bytes and its LRC checksum (8bit) if valid

        Notes
        -----
        Static overridden method
        """

        if not (isinstance(command, Command) and command.is_valid()):
            return None

        payload_to_build: bytearray = bytearray()

        payload_to_build.append(min(command.destination_logical_id, 0x64))
        payload_to_build.append(0x90)
        payload_to_build.append(0x02)

        # Set color, speed and brightness
        payload_to_build.append(0x01)
        payload_to_build.append(command.color)
        payload_to_build.append(command.blink_speed)
        payload_to_build.append(command.brightness)

        # Set duration
        payload_to_build.append(0x02)
        payload_to_build.append(0xFF)
        payload_to_build.append((command.duration_in_sec >> 8) & 0xFF)
        payload_to_build.append((command.duration_in_sec >> 0) & 0xFF)

        serialized_packet: dict[str, Any] = {
            "payload": bytes(payload_to_build),
            "checksum": utils.lrc8(bytes(payload_to_build)),
        }

        return common.BarePacket(**serialized_packet)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# MWings common defs

from __future__ import annotations

from struct import unpack
from enum import IntEnum, StrEnum, auto
from datetime import datetime, timezone, tzinfo
from collections import OrderedDict
from collections.abc import Sequence, MutableSequence, Iterable
from json import dumps
from abc import ABC, abstractmethod
from importlib import metadata
import platform
from typing import (
    Any,
    Callable,
    TypeVar,
    final,
    overload,
    TYPE_CHECKING,
)

from overrides import override
from pydantic_core import CoreSchema, core_schema
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    computed_field,
    field_serializer,
    field_validator,
)
from pydantic.types import AwareDatetime

from . import utils


try:
    import pandas as pd

    PANDAS_AVAILABLE = True
except ImportError:
    if TYPE_CHECKING:
        raise ImportError("pandas required for dev")
    else:
        PANDAS_AVAILABLE = False


Timezone: tzinfo = timezone.utc
"""Global tzinfo"""


SomeCallable = TypeVar("SomeCallable", bound=Callable[..., Any])
"""TypeVar for handlers"""


class DtypedDecimal(ABC):
    def get_dtype(self) -> str:
        """Provide dtype info for pandas

        Returns
        -------
        str
            dtype identifier
        """

        return "int64"

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Override the pydantic schema for serialization

        For JSON, serialize as plain int or float.
        For dict, serialize as object of subclass derived from int or float.

        Parameters
        ----------
        source_type : Any
            The class we are generating a schema for.
        handler : GetCoreSchemaHandler
            Call into Pydantic's internal schema generation.

        Returns
        -------
        CoreSchema
            A `pydantic-core` `CoreSchema`
        """

        return core_schema.json_or_python_schema(
            json_schema=core_schema.decimal_schema(),
            python_schema=core_schema.any_schema(),
        )


class UInt8(int, DtypedDecimal):
    def __new__(cls, value: int | None = None) -> UInt8:
        """Create immutable instance

        Parameters
        ----------
        value : int, optional
            Value to set

        Returns
        -------
        UInt8
            Wrapped value

        Raises
        ------
        ValueError
            Out of range
        """

        if value is None:
            value = 0
        if not (0 <= value < 2**8):
            raise ValueError("The given value is out of the range")
        return super(UInt8, cls).__new__(cls, value)

    @override
    def get_dtype(self) -> str:
        """Provide dtype info for pandas

        Returns
        -------
        str
            dtype identifier
        """

        return "uint8"

    def hex(self) -> str:
        """Returns hex representation

        python `hex()` function does not accept object of subclass
        derived from int.

        Returns
        -------
        str
            Hex string (lower case)
        """

        return f"{self:02x}"


class Int8(int, DtypedDecimal):
    def __new__(cls, value: int | None = None) -> Int8:
        """Create immutable instance

        Parameters
        ----------
        value : int, optional
            Value to set

        Returns
        -------
        Int8
            Wrapped value

        Raises
        ------
        ValueError
            Out of range
        """

        if value is None:
            value = 0
        if not (-(2**7) <= value < 2**7):
            raise ValueError("The given value is out of the range")
        return super(Int8, cls).__new__(cls, value)

    @override
    def get_dtype(self) -> str:
        """Provide dtype info for pandas

        Returns
        -------
        str
            dtype identifier
        """

        return "int8"

    def hex(self) -> str:
        """Returns hex representation

        python `hex()` function does not accept object of subclass
        derived from int.

        Returns
        -------
        str
            Hex string (lower case)
        """

        return f"{self:02x}"


class UInt16(int, DtypedDecimal):
    def __new__(cls, value: int | None = None) -> UInt16:
        """Create immutable instance

        Parameters
        ----------
        value : int, optional
            Value to set

        Returns
        -------
        UInt16
            Wrapped value

        Raises
        ------
        ValueError
            Out of range
        """

        if value is None:
            value = 0
        if not (0 <= value < 2**16):
            raise ValueError("The given value is out of the range")
        return super(UInt16, cls).__new__(cls, value)

    @override
    def get_dtype(self) -> str:
        """Provide dtype info for pandas

        Returns
        -------
        str
            dtype identifier
        """

        return "uint16"

    def hex(self) -> str:
        """Returns hex representation

        python `hex()` function does not accept object of subclass
        derived from int.

        Returns
        -------
        str
            Hex string (lower case)
        """

        return f"{self:04x}"


class Int16(int, DtypedDecimal):
    def __new__(cls, value: int | None = None) -> Int16:
        """Create immutable instance

        Parameters
        ----------
        value : int, optional
            Value to set

        Returns
        -------
        Int16
            Wrapped value

        Raises
        ------
        ValueError
            Out of range
        """

        if value is None:
            value = 0
        if not (-(2**15) <= value < 2**15):
            raise ValueError("The given value is out of the range")
        return super(Int16, cls).__new__(cls, value)

    @override
    def get_dtype(self) -> str:
        """Provide dtype info for pandas

        Returns
        -------
        str
            dtype identifier
        """

        return "int16"

    def hex(self) -> str:
        """Returns hex representation

        python `hex()` function does not accept object of subclass
        derived from int.

        Returns
        -------
        str
            Hex string (lower case)
        """

        return f"{self:04x}"


class UInt32(int, DtypedDecimal):
    def __new__(cls, value: int | None = None) -> UInt32:
        """Create immutable instance

        Parameters
        ----------
        value : int, optional
            Value to set

        Returns
        -------
        UInt32
            Wrapped value

        Raises
        ------
        ValueError
            Out of range
        """

        if value is None:
            value = 0
        if not (0 <= value < 2**32):
            raise ValueError("The given value is out of the range")
        return super(UInt32, cls).__new__(cls, value)

    @override
    def get_dtype(self) -> str:
        """Provide dtype info for pandas

        Returns
        -------
        str
            dtype identifier
        """

        return "uint32"

    def hex(self) -> str:
        """Returns hex representation

        python `hex()` function does not accept object of subclass
        derived from int.

        Returns
        -------
        str
            Hex string (lower case)
        """

        return f"{self:08x}"


class Int32(int, DtypedDecimal):
    def __new__(cls, value: int | None = None) -> Int32:
        """Create immutable instance

        Parameters
        ----------
        value : int, optional
            Value to set

        Returns
        -------
        Int32
            Wrapped value

        Raises
        ------
        ValueError
            Out of range
        """

        if value is None:
            value = 0
        if not (-(2**31) <= value < 2**31):
            raise ValueError("The given value is out of the range")
        return super(Int32, cls).__new__(cls, value)

    @override
    def get_dtype(self) -> str:
        """Provide dtype info for pandas

        Returns
        -------
        str
            dtype identifier
        """

        return "int32"

    def hex(self) -> str:
        """Returns hex representation

        python `hex()` function does not accept object of subclass
        derived from int.

        Returns
        -------
        str
            Hex string (lower case)
        """

        return f"{self:08x}"


class Float32(float, DtypedDecimal):
    def __new__(cls, value: float | None = None) -> Float32:
        if value is None:
            value = 0.0
        if not (-3.4028235e38 <= value <= 3.4028235e38):
            raise ValueError("The given value is out of the range")
        return super(Float32, cls).__new__(cls, value)

    @override
    def get_dtype(self) -> str:
        """Provide dtype info for pandas

        Returns
        -------
        str
            dtype identifier
        """

        return "float32"


class Float64(float, DtypedDecimal):
    def __new__(cls, value: float | None = None) -> Float64:
        if value is None:
            value = 0.0
        if not (-1.7976931348623157e308 <= value <= 1.7976931348623157e308):
            raise ValueError("The given value is out of the range")
        return super(Float64, cls).__new__(cls, value)

    @override
    def get_dtype(self) -> str:
        """Provide dtype info for pandas

        Returns
        -------
        str
            dtype identifier
        """

        return "float64"


T = TypeVar("T")
"""TypeVar for generics"""


class FixedList(MutableSequence[T]):
    """List with fixed length"""

    def __init__(
        self,
        length: int,
        initial_elements: Iterable[T],
    ):
        """Constructor for the sequence

        Parameters
        ----------
        length : int
            Fixed length
        initial_elements : Iterable[T]
            Initial items in iterable representation

        Raises
        ------
        ValueError
            Invalid length
        """

        if length <= 0:
            raise ValueError("Invalid length")
        if not initial_elements:
            raise ValueError(f"Expected {length} elements, got none")
        if length != sum(1 for e in initial_elements):
            raise ValueError(
                f"Expected {length} elements, got {sum(1 for e in initial_elements)}"
            )
        self.__length = length
        self.items = list(initial_elements)

    @overload
    def __getitem__(self, index: int) -> T:
        ...

    @overload
    def __getitem__(self, index: slice) -> MutableSequence[T]:
        ...

    @override
    def __getitem__(self, index: int | slice) -> Any:
        """Get item(s) with index

        Parameters
        ----------
        index : int | slice
            Index for getting data

        Returns
        -------
        Any
            Item or part of sequence

        Raises
        ------
        IndexError
            Out of range
        """

        match index:
            case int():
                if not (0 <= index < self.__length):
                    raise IndexError(
                        f"Index out of range, expected under {self.__length}"
                    )
            case slice():
                if not (
                    0 <= index.start < self.__length and 0 < index.stop <= self.__length
                ):
                    raise IndexError(
                        f"Index out of range, expected under {self.__length}"
                    )
        return self.items[index]

    @override
    def insert(self, index: int, value: T) -> None:
        """Insert an item for the specific index

        Parameters
        ----------
        index : int
            Index to insert
        value : T
            Value of item to insert

        Raises
        ------
        IndexError
            Out of range
        """

        if not (0 <= index < self.__length):
            raise IndexError(f"Index out of range, expected under {self.__length}")
        self.items[index] = value

    @overload
    def __setitem__(self, index: int, value: T) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[T]) -> None:
        ...

    @override
    def __setitem__(self, index: int | slice, value: Any) -> None:
        """Set item with index

        Parameters
        ----------
        index : int | slice
            Index to set
        value : Any
            Item or part of sequence to set

        Raises
        ------
        IndexError
            Out of range
        """

        match index:
            case int():
                self.insert(index, value)
            case slice():
                if not (
                    0 <= index.start < self.__length and 0 < index.stop <= self.__length
                ):
                    raise IndexError(
                        f"Index out of range, expected under {self.__length}"
                    )
                for i, v in enumerate(value):
                    self.insert(index.start + i, v)

    @overload
    def __delitem__(self, index: int) -> None:
        ...

    @overload
    def __delitem__(self, index: slice) -> None:
        ...

    @override
    def __delitem__(self, index: int | slice) -> None:
        """Delete item with index

        Parameters
        ----------
        index : int | slice
            Index to delete

        Raises
        ------
        RuntimeError
            FixedList does not support deletion
        """

        raise RuntimeError("FixedList does not support deletion")

    @override
    def __len__(self) -> int:
        """Get length

        Returns
        -------
        int
            Length of items
        """

        return self.__length

    @override
    def append(self, value: T) -> None:
        """Append an item

        Parameters
        ----------
        value : T
            Value of the item to append

        Raises
        ------
        IndexError
            FixedList does not support append()
        """

        raise IndexError("FixedList does not support append()")

    @override
    def extend(self, values: Iterable[T]) -> None:
        """Extend the sequence

        Parameters
        ----------
        values : Iterable[T]
            Sequence to extend

        Raises
        ------
        IndexError
            FixedList does not support extend()
        """

        raise IndexError("FixedList does not support extend()")

    @override
    def pop(self, index: int = -1) -> T:
        """Pop an item

        Parameters
        ----------
        index : int
            Index to pop

        Returns
        -------
        T
            An poped item

        Raises
        ------
        RuntimeError
            FixedList does not support pop()
        """

        raise RuntimeError("FixedList does not support pop()")

    @override
    def remove(self, value: T) -> None:
        """Remove an item

        Parameters
        ----------
        value : T
            Value of item to remove

        Raises
        ------
        RuntimeError
            FixedList does not support remove()
        """

        raise RuntimeError("FixedList does not support remove()")

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Override the pydantic schema for serialization

        For JSON, serialize as plain list.
        For dict, serialize as object of subclass derived from MutableSequence.

        Parameters
        ----------
        source_type : Any
            The class we are generating a schema for.
        handler : GetCoreSchemaHandler
            Call into Pydantic's internal schema generation.

        Returns
        -------
        CoreSchema
            A `pydantic-core` `CoreSchema`
        """

        return core_schema.json_or_python_schema(
            json_schema=core_schema.list_schema(),
            python_schema=core_schema.any_schema(),
        )


class FixedTuple(Sequence[T]):
    """Tuple with fixed length"""

    def __init__(self, length: int, elements: Iterable[T]):
        """Constructor for the sequence

        Parameters
        ----------
        length : int
            Fixed length
        elements : Iterable[T]
            Items to contain

        Raises
        ------
        ValueError
            Invalid length
        """

        if length <= 0:
            raise ValueError("Invalid length")
        if not elements:
            raise ValueError(f"Expected {length} elements, got none")
        if length != sum(1 for e in elements):
            raise ValueError(
                f"Expected {length} elements, got {sum(1 for e in elements)}"
            )
        self.__length = length
        self.items = tuple(elements)

    @overload
    def __getitem__(self, index: int) -> T:
        ...

    @overload
    def __getitem__(self, index: slice) -> MutableSequence[T]:
        ...

    @override
    def __getitem__(self, index: int | slice) -> Any:
        """Get item(s) with index

        Parameters
        ----------
        index : int | slice
            Index for getting data

        Returns
        -------
        Any
            Item or part of sequence to get
        """

        match index:
            case int():
                if not (0 <= index < self.__length):
                    raise IndexError(
                        f"Index out of range, expected under {self.__length}"
                    )
            case slice():
                if not (
                    0 <= index.start < self.__length and 0 < index.stop <= self.__length
                ):
                    raise IndexError(
                        f"Index out of range, expected under {self.__length}"
                    )
        return self.items[index]

    @override
    def __len__(self) -> int:
        """Get length

        Returns
        -------
        int
            Length for items
        """

        return self.__length

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Override the pydantic schema for serialization

        For JSON, serialize as plain tuple.
        For dict, serialize as object of subclass derived from Sequence.

        Parameters
        ----------
        source_type : Any
            The class we are generating a schema for.
        handler : GetCoreSchemaHandler
            Call into Pydantic's internal schema generation.

        Returns
        -------
        CoreSchema
            A `pydantic-core` `CoreSchema`
        """

        return core_schema.json_or_python_schema(
            json_schema=core_schema.tuple_variable_schema(),
            python_schema=core_schema.any_schema(),
        )


class CrossSectional(FixedTuple[T]):
    """Tuple for cross-sectional data such as ADC voltages"""

    pass


class TimeSeries(FixedTuple[T]):
    """Tuple for time series data such as acceleration samples"""

    pass


@final
class PacketType(StrEnum):
    """Packet type identifier for receiving handlers

    Attributes
    ----------
    BARE: str
        Identifier for bare packets
    ACT: str
        Identifier for act packets
    APP_TWELITE: str
        Identifier for App_Twelite packets
    APP_IO: str
        Identifier for App_IO packets
    APP_ARIA: str
        Identifier for App_ARIA packets
    APP_CUE: str
        Identifier for App_CUE packets
    APP_CUE_PAL_EVENT: str
        Identifier for App_CUE (PAL Move or Dice mode) packets
    APP_PAL_OPENCLOSE: str
        Identifier for App_PAL (OPENCLOSE) packets
    APP_PAL_AMB: str
        Identifier for App_PAL (AMB) packets
    APP_PAL_MOT: str
        Identifier for App_PAL (MOT) packets
    APP_UART_ASCII: str
        Identifier for App_Uart (Mode A) packets
    APP_UART_ASCII_EXTENDED: str
        Identifier for App_Uart (Mode A, Extended) packets
    """

    BARE = auto()
    ACT = auto()
    APP_TWELITE = auto()
    APP_IO = auto()
    APP_ARIA = auto()
    APP_CUE = auto()
    APP_CUE_PAL_EVENT = auto()
    APP_PAL_OPENCLOSE = auto()
    APP_PAL_AMB = auto()
    APP_PAL_MOT = auto()
    APP_UART_ASCII = auto()
    APP_UART_ASCII_EXTENDED = auto()


@final
class MagnetState(IntEnum):
    """Event ID for state of the magnet sensor

    For App_ARIA, App_CUE and App_PAL(OPENCLOSE)

    Attributes
    ----------
    NOT_DETECTED: int
        No magnet
    N_POLE_IS_CLOSE: int
        N pole is close
    S_POLE_IS_CLOSE: int
        S pole is close
    """

    NOT_DETECTED = 0x00
    N_POLE_IS_CLOSE = 0x01
    S_POLE_IS_CLOSE = 0x02


@final
class AccelEvent(IntEnum):
    """Event ID for state of the magnet sensor

    Attributes
    ----------
    DICE_1: int
        Dice roll: 1
    DICE_2: int
        Dice roll: 2
    DICE_3: int
        Dice roll: 3
    DICE_4: int
        Dice roll: 4
    DICE_5: int
        Dice roll: 5
    DICE_6: int
        Dice roll: 6
    SHAKE: int
        Shake
    MOVE: int
        Move
    NONE: int
        No events
    """

    DICE_1 = 0x01
    DICE_2 = 0x02
    DICE_3 = 0x03
    DICE_4 = 0x04
    DICE_5 = 0x05
    DICE_6 = 0x06
    SHAKE = 0x08
    MOVE = 0x10
    NONE = 0xFF


@final
class AppPalNoticeColor(IntEnum):
    """Color ID for App_PAL (NOTICE)

    Attributes
    ----------
    RED : int
        RED color
    GREEN : int
        GREEN color
    BLUE : int
        BLUE color
    YELLOW : int
        YELLOW color
    PURPLE : int
        PURPLE color
    LIGHT_BLUE : int
        LIGHT_BLUE color
    WHITE : int
        WHITE color
    WARM_WHITE : int
        WARM_WHITE color
    """

    RED = 0
    GREEN = 1
    BLUE = 2
    YELLOW = 3
    PURPLE = 4
    LIGHT_BLUE = 5
    WHITE = 6
    WARM_WHITE = 7


@final
class AppPalNoticeBlinkSpeed(IntEnum):
    """Blinking speed ID for App_PAL (NOTICE)

    Attributes
    ----------
    ALWAYS_ON : int
        Always on
    SLOW : int
        Slow blinking
    MEDIUM : int
        Medium blinking
    FAST : int
        Fast blinking
    """

    ALWAYS_ON = 0
    SLOW = 1
    MEDIUM = 2
    FAST = 3


@final
class AppPalNoticeRGBWColor(BaseModel):
    """Color in RGBW for App_PAL (NOTICE)

    Attributes
    ----------
    red : UInt8
        Red value 0-0xF
    green : UInt8
        Green value 0-0xF
    blue : UInt8
        Blue value 0-0xF
    white : UInt8
        White value 0-0xF
    """

    red: UInt8 = Field(default=UInt8(0), ge=UInt8(0), le=UInt8(0xF))
    green: UInt8 = Field(default=UInt8(0), ge=UInt8(0), le=UInt8(0xF))
    blue: UInt8 = Field(default=UInt8(0), ge=UInt8(0), le=UInt8(0xF))
    white: UInt8 = Field(default=UInt8(0xF), ge=UInt8(0), le=UInt8(0xF))

    def u16(self) -> UInt16:
        """Returns UInt16 representation

        Returns
        -------
        UInt16
            RGBW as UInt16
        """

        return UInt16(
            (self.white & 0xF) << 12
            | (self.blue & 0xF) << 8
            | (self.green & 0xF) << 4
            | (self.red & 0xF) << 0
        )


@final
class BarePacket(BaseModel):
    """Bare packet dataclass

    Attributes
    ----------
    payload: bytes
        Data payload
    checksum: UInt8
        LRC checksum for data payload
    """

    payload: bytes
    checksum: UInt8

    def __init__(
        self,
        payload: bytes,
        checksum: UInt8 | None = None,
        logical_and_command_id: tuple[UInt8, UInt8] | None = None,
    ):
        """Overridden constructor

        Parameters
        ----------
        payload : bytes
            Payload data
        checksum : UInt8, optional
            LRC8 checksum
        logical_and_command_id : tuple[UInt8, UInt8], optional
            Logical ID and Command ID (if set, payload should be
            shorter
        """

        if logical_and_command_id is not None:
            full_payload_data = bytearray(payload)
            full_payload_data[0:0] = bytes(logical_and_command_id)
            full_payload = bytes(full_payload_data)
            if checksum is not None:
                super().__init__(payload=full_payload, checksum=checksum)
            else:
                super().__init__(
                    payload=full_payload, checksum=utils.lrc8(full_payload)
                )
        else:
            if checksum is not None:
                super().__init__(payload=payload, checksum=checksum)
            else:
                super().__init__(payload=payload, checksum=utils.lrc8(payload))

    def u8_from(self, index: int) -> bytes | None:
        """Get bytes from the specified position in the payload

        Parameters
        ----------
        index : int
            Position index

        Returns
        -------
        bytes | None
            return data if valid else None
        """
        return bytes(self.payload[index:]) if index < len(self.payload) else None

    def u8_at(self, index: int) -> UInt8:
        """Get 1 byte as an unsigned integer for the specified position in the payload

        Parameters
        ----------
        index : int
            Position index

        Returns
        -------
        UInt8
            return data if valid else zero
        """
        return UInt8(self.payload[index] if index < len(self.payload) else 0)

    def i8_at(self, index: int) -> Int8:
        """Get 1 byte as a signed integer for the specified position in the payload

        Parameters
        ----------
        index : int
            Position index

        Returns
        -------
        Int8
            return data if valid else zero
        """
        return Int8(
            int(unpack(">b", self.payload[index : index + 1])[0])
            if index < len(self.payload)
            else 0
        )

    def u16_at(self, index: int) -> UInt16:
        """Get 2 bytes as an unsigned integer for the specified position in the payload

        Parameters
        ----------
        index : int
            Position index

        Returns
        -------
        UInt16
            return data if valid else zero
        """
        return UInt16(
            int(unpack(">H", self.payload[index : index + 2])[0])
            if index + 1 < len(self.payload)
            else 0
        )

    def i16_at(self, index: int) -> Int16:
        """Get 2 bytes as an signed integer for the specified position in the payload

        Parameters
        ----------
        index : int
            Position index

        Returns
        -------
        Int16
            return data if valid else zero
        """
        return Int16(
            int(unpack(">h", self.payload[index : index + 2])[0])
            if index + 1 < len(self.payload)
            else 0
        )

    def u32_at(self, index: int) -> UInt32:
        """Get 4 bytes as an unsigned integer for the specified position in the payload

        Parameters
        ----------
        index : int
            Position index

        Returns
        -------
        UInt32
            return data if valid else zero
        """
        return UInt32(
            int(unpack(">I", self.payload[index : index + 4])[0])
            if index + 3 < len(self.payload)
            else 0
        )

    def i32_at(self, index: int) -> Int32:
        """Get 4 bytes as an signed integer for the specified position in the payload

        Parameters
        ----------
        index : int
            Position index

        Returns
        -------
        Int32
            return data if valid else zero
        """
        return Int32(
            int(unpack(">i", self.payload[index : index + 4])[0])
            if index + 3 < len(self.payload)
            else 0
        )


class ParsedPacketBase(ABC, BaseModel):
    """Base dataclass for data of parsed packets

    Attributes
    ----------
    mwings_implementation: str
        Implementation of mwings; In this case: "python"
    mwings_version: str
        Version of mwings in PEP440 format declared in the pyproject.toml
    time_parsed: str | None
        Date and time parsed in ISO 8601 format
    hostname: str
        Hostname for the running system
    system_type: str
        System type for the running system (e.g. "Linux")
    packet_type: PacketType
        Type of the received packet
    sequence_number: UInt16
        Sequence number for the packet
    source_serial_id: UInt32
        Serial ID for the source device
    source_logical_id: UInt8
        Logical ID for the source device
    lqi: UInt8
        Link quality indicator for the source device (Max: 255)
    supply_voltage: UInt16
        Supply voltage for the source device in mV

    Notes
    -----
    Immutable (frozen) object
    """

    model_config = ConfigDict(frozen=True)

    time_parsed: AwareDatetime | None = Field(default=None)

    packet_type: PacketType = Field(default=PacketType.BARE)
    sequence_number: UInt16 | None = Field(default=None, ge=0, le=0xFFFF)
    source_serial_id: UInt32 = Field(
        default=UInt32(0), ge=UInt32(0), le=UInt32(0xFFFFFFFF)
    )
    source_logical_id: UInt8 = Field(default=UInt8(0))
    lqi: UInt8 | None = Field(default=None, ge=UInt8(0), le=UInt8(255))
    supply_voltage: UInt16 | None = Field(default=None, ge=UInt16(0), le=UInt16(0xFFFF))

    @computed_field
    def mwings_implementation(self) -> str:
        return "python"

    @computed_field
    def mwings_version(self) -> str:
        return metadata.version(__name__.split(".")[0])

    @computed_field
    def hostname(self) -> str:
        return platform.node()

    @computed_field
    def system_type(self) -> str:
        return platform.system()

    @field_validator("time_parsed")
    @classmethod
    def datetime_must_be_clear(cls, dt: datetime | None) -> datetime | None:
        """Check time received

        Must be aware timezone as mw.common.Timezone

        Parameters
        ----------
        dt : datetime
            Input

        Returns
        -------
        datetime
            Valid input

        Raises
        ------
        ValueError
            Native or not same as mw.common.Timezone
        """

        if dt is not None and dt.tzinfo is not Timezone:
            raise ValueError("datetime must be aware and same as mw.common.Timezone.")
        return dt

    @field_validator("source_logical_id")
    @classmethod
    def check_source_logical_id(cls, lid: UInt8) -> UInt8:
        """Check source logical id

        Must be in range between 0 and 100 or 120 and 127

        Parameters
        ----------
        lid : UInt8
            Input

        Returns
        -------
        UInt8
            Valid input

        Raises
        ------
        ValueError
            Out of range
        """

        if 0 <= lid <= 100:
            return lid
        elif lid in range(120, 128):
            return lid
        elif lid in (0xFE, 0xFF):
            return lid
        else:
            raise ValueError("must be in range (0-100) or (120-127)")

    @field_serializer("time_parsed")
    def serialize_time_parsed(self, dt: datetime | None) -> str | None:
        """Print time_parsed in ISO 8601 format

        Parameters
        ----------
        dt : datetime, optional
            Date and time received

        Returns
        -------
        str | None
            Serialized text for JSON or something

        Notes
        -----
        Date and time should be UTC, but python uses "+00:00" suffix instead of "Z".
        However, it can be parsed in other environments like ECMAScript's Date().
        """
        if dt is None:
            return None
        return dt.isoformat()  # YYYY-MM-DDThh:mm:ss.ssssss+00:00

    @field_serializer("packet_type")
    def serialize_packet_type(self, packet_type: PacketType) -> str:
        """Print packet_type in readable names for JSON or something

        Parameters
        ----------
        packet_type : PacketType
            Type of the packet

        Returns
        -------
        str
            Serialized text for JSON or something
        """
        return packet_type.name

    @field_serializer("source_serial_id")
    def serialize_source_serial_id(self, source_serial_id: UInt32) -> str:
        """Print source_serial_id in HEX for JSON or something

        Parameters
        ----------
        source_serial_id : UInt32
            Source serial ID

        Returns
        -------
        str
            Serialized text for JSON or something
        """

        return source_serial_id.hex().upper()

    def to_dict(
        self,
        include: set[str] | None = None,
        exclude: set[str] | None = None,
        verbose: bool = True,
        spread: bool = False,
        sort_keys: bool = False,
    ) -> dict[str, Any]:
        """Export to a dictionary (or OrderedDict)

        Parameters
        ----------
        include : set[str], optional
            Properties to include in the dictionary
        exclude : set[str], optional
            Properties to exclude in the dictionary
        verbose : bool
            Set False to exclude system information.
            Only valid if include and exclude are None.
        spread : bool
            Spread cross-sectional tuple values into separated properties if True
        sort_keys : bool
            Returns a sorted OrderedDict if True

        Returns
        -------
        dict[str, Any]
            Output dictionary (or OrderedDict)

        Notes
        -----
        Higher-level implementation of pydantic's model_dump()
        """

        if not spread:
            if not sort_keys:
                if not verbose and not any((include, exclude)):
                    return self.model_dump(
                        exclude={
                            "mwings_implementation",
                            "mwings_version",
                            "system_type",
                            "hostname",
                        }
                    )
                else:
                    return self.model_dump(include=include, exclude=exclude)
            else:
                return OrderedDict(
                    sorted(
                        self.to_dict(
                            include=include, exclude=exclude, verbose=verbose
                        ).items()
                    )
                )
        else:
            ordered_dict = OrderedDict()
            for key, value in self.to_dict(
                include=include, exclude=exclude, verbose=verbose, sort_keys=sort_keys
            ).items():
                if isinstance(value, CrossSectional):
                    for tup_index, tup_elem in enumerate(value):
                        ordered_dict[f"{key}_{tup_index+1}"] = tup_elem
                else:
                    ordered_dict[key] = value
            return ordered_dict if sort_keys else dict(ordered_dict)

    def to_json(
        self,
        include: set[str] | None = None,
        exclude: set[str] | None = None,
        verbose: bool = True,
        spread: bool = False,
        indent: int | None = 2,
        sort_keys: bool = False,
    ) -> str:
        """Export to a JSON string

        Parameters
        ----------
        include : set[str], optional
            Properties to include the JSON
        exclude : set[str], optional
            Properties to exclude the JSON
        verbose : bool
            Set False to exclude system information.
            Only valid if include and exclude are None.
        spread : bool
            Spread cross-sectional tuple values into separated properties if True
        indent : int, optional
            Space-indentation width (default: 2, None to single-line)
        sort_keys : bool
            Sort properties by keys if True

        Returns
        -------
        str
            Output JSON
        """

        dic = self.to_dict(
            include=include,
            exclude=exclude,
            verbose=verbose,
            spread=spread,
            sort_keys=sort_keys,
        )
        for key, value in dic.items():
            if isinstance(value, FixedTuple):
                dic[key] = tuple(value)  # Make serializable
        return dumps(dic, indent=indent)

    def to_df(
        self,
        include: set[str] | None = None,
        exclude: set[str] | None = None,
        verbose: bool = True,
    ) -> pd.DataFrame:
        """Export to a pandas DataFrame

        Requires optional dependency: pandas

        Parameters
        ----------
        include : set[str], optional
            Columns to include in the DataFrame
        exclude : set[str], optional
            Columns to include in the DataFrame
        verbose : bool
            Set False to exclude system information.
            Only valid if include and exclude are None.

        Returns
        -------
        pd.DataFrame
            Output DataFrame

        Raises
        ------
        EnvironmentError
            No pandas found
        """

        if PANDAS_AVAILABLE:
            dic = self.to_dict(
                include=include,
                exclude=exclude,
                verbose=verbose,
                spread=True,
            )
            df = pd.DataFrame()
            time_series_max = max(
                [
                    len(value) if isinstance(value, TimeSeries) else 0
                    for value in dic.values()
                ]
            )
            if time_series_max == 0:
                for key in dic:
                    df[key] = [dic[key]]
                    match dic[key]:
                        case DtypedDecimal():
                            df[key] = df[key].astype(dic[key].get_dtype())
                        case str():
                            df[key] = df[key].astype("string")
                df.insert(1, "time_series", [0])
            else:
                for key in dic:
                    if isinstance(dic[key], TimeSeries):
                        df[key] = [
                            dic[key][i] if dic[key][i] else type(dic[key][0])()
                            for i in range(time_series_max)
                        ]
                        match dic[key][0]:
                            case DtypedDecimal():
                                df[key] = df[key].astype(dic[key][0].get_dtype())
                            case str():
                                df[key] = df[key].astype("string")
                    else:
                        df[key] = [dic[key] for _ in range(time_series_max)]
                        match dic[key]:
                            case DtypedDecimal():
                                df[key] = df[key].astype(dic[key].get_dtype())
                            case str():
                                df[key] = df[key].astype("string")
                df.insert(1, "time_series", range(time_series_max))
            df["time_series"] = df["time_series"].astype("uint8")
            df["time_parsed"] = df["time_parsed"].apply(
                pd.to_datetime, utc=Timezone is timezone.utc, format="ISO8601"
            )
            return df
        else:
            raise EnvironmentError("to_df() requires pandas")


SomeParsedPacket = TypeVar("SomeParsedPacket", bound=ParsedPacketBase)
"""TypeVar for all classes derived from ParsedPacketBase"""


class CommandBase(ABC, BaseModel):
    """Base dataclass for commands

    Attributes
    ----------
    destination_logical_id: UInt8
        Logical ID for the source device
    """

    model_config = ConfigDict()

    destination_logical_id: UInt8 = Field(default=UInt8(0x78))

    @abstractmethod
    def is_valid(self) -> bool:
        """Check if the command content is valid or not

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Pure virtual function
        """
        pass


class PacketParserBase(ABC):
    """Base class for packet parsers"""

    @staticmethod
    @abstractmethod
    def is_valid(bare_packet: BarePacket) -> bool:
        """Check if the given bare packet is valid or not

        Parameters
        ----------
        bare_packet : BarePacket
            Bare packet content

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Pure virtual function
        """
        pass

    @staticmethod
    @abstractmethod
    def parse(bare_packet: BarePacket) -> ParsedPacketBase | None:
        """Parse the given bare packet

        Parameters
        ----------
        bare_packet : BarePacket
            Bare packet content

        Returns
        -------
        ParsedPacketBase | None
            Parsed packet content if valid otherwise None

        Notes
        -----
        Pure virtual function
        """
        pass


SomeCommand = TypeVar("SomeCommand", bound=CommandBase)
"""TypeVar for all classes derived from CommandBase"""


class CommandSerializerBase(ABC):
    """Base class for packet serializers"""

    @staticmethod
    @abstractmethod
    def serialize(command: SomeCommand) -> BarePacket | None:
        """Serialize the given command

        Parameters
        ----------
        command : SomeCommand
            Some command to serialize

        Returns
        -------
        BarePacket | None
            Serialized bytes and its LRC checksum (8bit) if valid
        """
        pass



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet parser for App_CUE (CUE mode)

from datetime import datetime
from typing import Any, final

from overrides import override
from pydantic import Field, field_serializer

from .. import common


@final
class ParsedPacket(common.ParsedPacketBase):
    """Dataclass for parsed packets from App_CUE

    Attributes
    ----------
    router_serial_id: UInt32
        Serial ID for the first router device (0x80000000 with no routing)
    ai1_voltage: common.UInt16
        Voltage for AI1 port in mV
    sample_count: common.UInt8
        Number of accel samples
    samples_x: common.TimeSeries[common.Int16]
        Accel samples for x axis
    samples_y: common.TimeSeries[common.Int16]
        Accel samples for y axis
    samples_z: common.TimeSeries[common.Int16]
        Accel samples for z axis
    has_accel_event: bool
        True if an accel event is available
    accel_event: common.AccelEvent
        Accel event
    magnet_state: common.MagnetState
        Magnet state
    magnet_state_changed: bool
        True if the magnet state was changed
    """

    router_serial_id: common.UInt32 = Field(
        default=common.UInt32(0), ge=common.UInt32(0), le=common.UInt32(0xFFFFFFFF)
    )
    ai1_voltage: common.UInt16 = Field(
        default=common.UInt16(0), ge=common.UInt16(0), le=common.UInt16(3600)
    )
    sample_count: common.UInt8 = Field(
        default=common.UInt8(10), ge=common.UInt8(10), le=common.UInt8(10)
    )
    samples_x: common.TimeSeries[common.Int16] = Field(
        default=common.TimeSeries[common.Int16](
            10, [common.Int16(0) for _ in range(10)]
        )
    )
    samples_y: common.TimeSeries[common.Int16] = Field(
        default=common.TimeSeries[common.Int16](
            10, [common.Int16(0) for _ in range(10)]
        )
    )
    samples_z: common.TimeSeries[common.Int16] = Field(
        default=common.TimeSeries[common.Int16](
            10, [common.Int16(0) for _ in range(10)]
        )
    )
    has_accel_event: bool = Field(default=False)
    accel_event: common.AccelEvent = Field(default=common.AccelEvent.NONE)
    magnet_state: common.MagnetState = Field(default=common.MagnetState.NOT_DETECTED)
    magnet_state_changed: bool = Field(default=False)

    @field_serializer("accel_event")
    def serialize_accel_event(self, accel_event: common.AccelEvent) -> str:
        """Print accel_event in readable names for JSON or something

        Parameters
        ----------
        accel_event : common.AccelEvent
            Accel event

        Returns
        -------
        str
            Serialized text for JSON or something
        """

        return accel_event.name

    @field_serializer("magnet_state")
    def serialize_magnet_state(self, magnet_state: common.MagnetState) -> str:
        """Print magnet_state in readable names for JSON or something

        Parameters
        ----------
        magnet_state : common.MagnetState
            Magnet state

        Returns
        -------
        str
            Serialized text for JSON or something
        """

        return magnet_state.name


@final
class PacketParser(common.PacketParserBase):
    """Packet parser for App_CUE"""

    @staticmethod
    @override
    def is_valid(bare_packet: common.BarePacket) -> bool:
        """Check the given bare packet is valid or not

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Static overridden method
        """
        if (
            (bare_packet.u8_at(0) & 0x80) == 0x80
            and (bare_packet.u8_at(7) & 0x80) == 0x80
            and bare_packet.u8_at(12) == 0x80
            and bare_packet.u8_at(13) == 0x05
            and len(bare_packet.payload) == 148
        ):
            return True
        return False

    @staticmethod
    @override
    def parse(bare_packet: common.BarePacket) -> ParsedPacket | None:
        """Try to parse the given bare packet

        Parameters
        ----------
        bare_packet : common.BarePacket
            Bare packet content

        Returns
        -------
        ParsedPacket | None
            Parsed packet data if valid else None

        Notes
        -----
        Static overridden method
        """
        if not PacketParser.is_valid(bare_packet):
            return None
        parsed_packet_data: dict[str, Any] = {
            "time_parsed": datetime.now(common.Timezone),
            "packet_type": common.PacketType.APP_CUE,
            "sequence_number": bare_packet.u16_at(5),
            "source_serial_id": bare_packet.u32_at(7),
            "source_logical_id": bare_packet.u8_at(11),
            "lqi": bare_packet.u8_at(4),
            "supply_voltage": bare_packet.u16_at(34),
            "router_serial_id": bare_packet.u32_at(0),
            "ai1_voltage": bare_packet.u16_at(40),
            "sample_count": 10,
            "samples_x": common.TimeSeries[common.Int16](
                10,
                [bare_packet.i16_at(51 + (10 * index) + 0) for index in range(10)],
            ),
            "samples_y": common.TimeSeries[common.Int16](
                10,
                [bare_packet.i16_at(51 + (10 * index) + 2) for index in range(10)],
            ),
            "samples_z": common.TimeSeries[common.Int16](
                10,
                [bare_packet.i16_at(51 + (10 * index) + 4) for index in range(10)],
            ),
            "has_accel_event": bool(bare_packet.u8_at(24) == 0x04),
            "accel_event": common.AccelEvent(
                bare_packet.u8_at(26) if bare_packet.u8_at(24) == 0x04 else 0xFF
            ),
            "magnet_state": common.MagnetState(bare_packet.u8_at(46) & 0x0F),
            "magnet_state_changed": False if bare_packet.u8_at(46) & 0x80 else True,
        }
        return ParsedPacket(**parsed_packet_data)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet serializer for App_Twelite

from typing import Any, final

from pydantic import Field
from overrides import override

from .. import common
from .. import utils


@final
class Command(common.CommandBase):
    """Dataclass for App_Twelite command

    Attributes
    ----------
    di_to_change: common.FixedList[bool]
        To enable modification on a specific digital interface, set True
    di_state: common.FixedList[bool]
        Output status for each digital interfaces
    pwm_to_change: common.FixedList[bool]
        To enable modification on a specific PWM interface, set True
    pwm_duty: common.FixedList[bool]
        Duty for each PWM interfaces (0 to 1024, can be disabled with 0xFFFF)
    """

    di_to_change: common.FixedList[bool] = Field(
        default=common.FixedList[bool](4, [True, True, True, True])
    )
    di_state: common.FixedList[bool] = Field(
        default=common.FixedList[bool](4, [False, False, False, False])
    )
    pwm_to_change: common.FixedList[bool] = Field(
        default=common.FixedList[bool](4, [True, True, True, True])
    )
    pwm_duty: common.FixedList[int] = Field(
        default=common.FixedList[int](4, [0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF])
    )

    @override
    def is_valid(self) -> bool:
        """Check if the command content is valid or not

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Overridden
        """

        return 0x00 <= self.destination_logical_id <= 0x78 and all(
            duty <= 1024 or duty == 0xFFFF for duty in self.pwm_duty
        )


@final
class CommandSerializer(common.CommandSerializerBase):
    """Command serializer for App_Twelite"""

    @staticmethod
    @override
    def serialize(command: common.SomeCommand) -> common.BarePacket | None:
        """Serialize the given command

        Parameters
        ----------
        command : common.SomeCommand
            App_Twelite command to serialize

        Returns
        -------
        common.BarePacket | None
            Serialized bytes and its LRC checksum (8bit) if valid

        Notes
        -----
        Static overridden method
        """

        if not (isinstance(command, Command) and command.is_valid()):
            return None

        payload_to_build: bytearray = bytearray()

        payload_to_build.append(min(command.destination_logical_id, 0x78))
        payload_to_build.append(0x80)
        payload_to_build.append(0x01)
        payload_to_build.append(
            sum((1 if command.di_state[port] else 0) << port for port in range(4))
        )
        payload_to_build.append(
            sum((1 if command.di_to_change[port] else 0) << port for port in range(4))
        )
        for port in range(4):
            if not (command.pwm_to_change[port]) or command.pwm_duty[port] == 0xFFFF:
                payload_to_build.extend([0xFF, 0xFF])
            else:
                payload_to_build.extend(
                    [
                        (min(command.pwm_duty[port], 1024) >> 8) & 0xFF,
                        (min(command.pwm_duty[port], 1024) >> 0) & 0xFF,
                    ]
                )

        serialized_packet: dict[str, Any] = {
            "payload": bytes(payload_to_build),
            "checksum": utils.lrc8(bytes(payload_to_build)),
        }

        return common.BarePacket(**serialized_packet)



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# MWings serializers

# Redundant imports for mypy
from . import app_twelite as app_twelite
from . import app_io as app_io
from . import app_pal_notice as app_pal_notice
from . import app_pal_notice_detailed as app_pal_notice_detailed
from . import app_pal_notice_event as app_pal_notice_event
from . import app_uart_ascii as app_uart_ascii



# -*- coding:utf-8 -*-
# Written for Python 3.12
# Formatted with Black

# Packet serializer for App_UART (Mode A) commands

from typing import final

from pydantic import Field
from overrides import override

from .. import common


@final
class Command(common.CommandBase):
    """Dataclass for App_UART (Mode A) command

    Attributes
    ----------
    command_id : common.UInt8
        Command id
    data : bytes
        Data to send
    """

    command_id: common.UInt8 = Field(
        default=common.UInt8(0), ge=common.UInt8(0), lt=common.UInt8(0x80)
    )
    data: bytes = Field(min_length=1, max_length=80)

    @override
    def is_valid(self) -> bool:
        """Check if the command content is valid or not

        Returns
        -------
        bool
            True if valid

        Notes
        -----
        Overridden
        """

        return (
            0x00 <= self.destination_logical_id <= 0x64
            or self.destination_logical_id == 0x78
        )


@final
class CommandSerializer(common.CommandSerializerBase):
    """Command serializer for App_Uart (Mode A)"""

    @staticmethod
    @override
    def serialize(command: common.SomeCommand) -> common.BarePacket | None:
        """Serialize the given command

        Parameters
        ----------
        command : common.SomeCommand
            App_Uart (Mode A) command to serialize

        Returns
        -------
        common.BarePacket | None
            Serialized bytes and its LRC checksum (8bit) if valid

        Notes
        -----
        Static overridden method
        """

        if not (isinstance(command, Command) and command.is_valid()):
            return None

        return common.BarePacket(
            payload=command.data,
            logical_and_command_id=(command.destination_logical_id, command.command_id),
        )
